from collections import defaultdict
from copy import deepcopy
from operator import itemgetter

from ..audioinfo import FILENAME
from ..constants import VARIOUS
from ..findfunc import filenametotag
from ..puddleobjects import natural_sort_key, ratio
from ..tagsources import RetrievalError
from ..translations import translate
from ..util import sorted_split_by_field, split_by_field, to_string
from ..mainwin.tagsources import (strip as strip_fields,
                                  DEFAULT_REGEXP,
                                  apply_regexps)


def set_status(v):
    print(v)


NO_MATCH_OPTIONS = [
    translate('Masstagging', 'Continue'),
    translate('Masstagging', 'Stop')]

SINGLE_MATCH_OPTIONS = [
    translate('Masstagging', 'Combine and continue'),
    translate('Masstagging', 'Replace and continue'),
    translate('Masstagging', 'Combine and stop'),
    translate('Masstagging', 'Replace and stop')]

AMBIGIOUS_MATCH_OPTIONS = [
    translate('Masstagging', 'Use best match'),
    translate('Masstagging', 'Do nothing and continue')]

COMBINE_CONTINUE = 0
REPLACE_CONTINUE = 1
COMBINE_STOP = 2
REPLACE_STOP = 3

CONTINUE = 0
STOP = 1

USE_BEST = 0
DO_NOTHING = 1
RETRY = 2

ALBUM_BOUND = 'album'
TRACK_BOUND = 'track'
PATTERN = 'pattern'
SOURCE_CONFIGS = 'source_configs'
FIELDS = 'fields'
JFDI = 'jfdi'
NAME = 'name'
DESC = 'description'
EXISTING_ONLY = 'field_exists'

DEFAULT_PATTERN = '%artist% - %album%/%track% - %title%'
DEFAULT_NAME = translate('Masstagging', 'Default Profile')

POLLING = translate("Masstagging", "<b>Polling: {}</b>")
MATCH_ARTIST_ALBUM = translate("Masstagging",
                               'Retrieving matching album. <b>{} - {}</b>')
MATCH_ARTIST = translate("Masstagging",
                         'Retrieving matching album. Artist=<b>{}</b>')
MATCH_ALBUM = translate("Masstagging",
                        'Retrieving matching album. Album=<b>{}</b>')
MATCH_NO_INFO = translate("Masstagging", 'Retrieving matching album.')

SEARCHING_ARTIST_ALBUM = ':insert' + translate("Masstagging",
                                               'Starting search for: <br />artist=<b>{}</b> '
                                               '<br />album=<b>{}</b><br />')
SEARCHING_ARTIST = ':insert' + translate("Masstagging",
                                         'Starting search for: <br />artist=<b>{}</b>'
                                         '<br />album=No album name found.')
SEARCHING_ALBUM = ':insert' + translate("Masstagging",
                                        'Starting search for: <br />album=<b>{}</b>'
                                        '<br />artist=No artist found.')
SEARCHING_NO_INFO = ':insert' + translate("Masstagging",
                                          'No artist or album info found in files. Starting search.')

RESULTS_FOUND = translate("Masstagging", '<b>{}</b> results found.')
NO_RESULTS_FOUND = translate("Masstagging", '<b>No results were found.</b>')
ONE_RESULT_FOUND = translate("Masstagging", '<b>One</b> result found.')

MATCHING_ALBUMS_FOUND = translate("Masstagging",
                                  "<b>{}</b> possibly matching albums found.")
ONE_MATCHING_ALBUM_FOUND = translate("Masstagging",
                                     '<b>One</b> possibly matching album found.')
NO_MATCHES = translate("Masstagging",
                       "No matches found for tag source <b>{}</b>")

RETRIEVING_NEXT = translate("Masstagging",
                            'Previously retrieved result does not match. '
                            'Retrieving next matching album.')

RECHECKING = translate("Masstagging",
                       '<br />Rechecking with results from <b>{}</b>.<br />')

VALID_FOUND = translate("Masstagging",
                        '<br />Valid matches were found for the album.')

NO_VALID_FOUND = translate("Masstagging",
                           '<b>No valid matches were found for the album.</b>')


class MassTagFlag(object):

    def __init__(self):
        self.stop = False
        object.__init__(self)


def brute_force_results(audios, retrieved):
    matched = {}

    audios = sorted(audios, key=lambda f: natural_sort_key(to_string(f.get('track', f['__filename']))))
    retrieved = sorted(retrieved, key=lambda t: natural_sort_key(to_string(t.get('track', t.get('title', '')))))

    for audio, result in zip(audios, retrieved):
        matched[audio] = result

    return matched


def check_result(result, audios):
    track_nums = [_f for _f in [to_string(audio.get('track', None)) for audio in audios] if _f]

    if result.tracks is None:
        return True

    if track_nums:
        max_num = 0
        for num in track_nums:
            try:
                num = int(num)
            except (TypeError, ValueError):
                continue
            max_num = num if num > max_num else max_num
        if max_num != 0 and max_num == len(result.tracks):
            return True

    if len(audios) == len(result.tracks):
        return True
    return False


def combine_tracks(track1, track2, repl=None):
    ret = defaultdict(lambda: [])

    for key, value in list(track2.items()) + list(track1.items()):
        if isinstance(value, str):
            if value not in ret[key]:
                ret[key].append(value)
        else:
            for v in value:
                if v not in ret[key]:
                    ret[key].append(v)
    if repl:
        for key in repl:
            if key in track2:
                ret[key] = track2[key]
    if '#exact' in track1:
        ret['#exact'] = track1['#exact']
    elif '#exact' in track2:
        ret['#exact'] = track2['#exact']
    return ret


def fields_from_text(text):
    if not text:
        return []
    return [_f for _f in map(str.strip, text.split(',')) if _f]


def dict_difference(dict1, dict2):
    """Returns a dictonary containing key/value pairs from dict2 where key
    isn't in dict1."""
    temp = {}
    for field in dict2:
        if field not in dict1:
            temp[field] = dict2[field]
    return temp


def find_best(matches, files, minimum=0.7):
    group = split_by_field(files, 'album', 'artist')
    album = list(group.keys())[0]
    artists = list(group[album].keys())
    if len(artists) == 1:
        artist = artists[0]
    else:
        artist = VARIOUS
    d = {'artist': artist, 'album': album}
    scores = {}

    for match in matches:
        if hasattr(match, 'info'):
            info = match.info
        else:
            info = match[0]

        score = min([ratio_compare(d, info, key) for key in d])

        if score in scores:
            score = score + 0.01  # For albums that have same name.
        scores[score] = match
        tracks = match.tracks if hasattr(match, 'tracks') else match[1]
        if tracks and score < minimum:
            if len(tracks) == len(files):
                scores[minimum + 0.01] = match

    if scores:
        return [scores[z] for z in
                sorted(scores, reverse=True) if z >= minimum]
    else:
        return []


def get_artist_album(files):
    tags = split_by_field(files, 'album', 'artist')
    album = list(tags.keys())[0]
    artists = tags[album]
    if len(artists) > 1:
        return VARIOUS, album
    else:
        return list(artists)[0], album


def get_match_str(info):
    artist = album = None
    if info.get('artist'):
        artist = to_string(info['artist'])

    if info.get('album'):
        album = to_string(info['album'])

    if artist and album:
        return MATCH_ARTIST_ALBUM.format(artist, album)
    elif artist:
        return MATCH_ARTIST.format(artist)
    elif album:
        return MATCH_ALBUM.format(album)
    else:
        return MATCH_NO_INFO


get_lower = lambda f, key, default = '': to_string(f.get(key, default)).lower()


def ratio_compare(d1, d2, key):
    return ratio(get_lower(d1, key, 'a'), get_lower(d2, key, 'b'))


def match_files(files, tracks, minimum=0.7, keys=None, jfdi=False,
                existing=False, as_index=False):
    if not keys:
        keys = ['artist', 'title']
    if 'track' in keys and len(keys) > 1:
        keys = keys[::]
        keys.remove('track')
    ret = {}
    replace_tracknumbers(files, tracks)
    assigned = {}
    matched = defaultdict(lambda: {})
    b = False

    for f_index, f in enumerate(files):
        scores = {}
        for t_index, track in enumerate(tracks):
            totals = [ratio_compare(f, track, key) for key in keys]
            if not totals:
                continue
            score = min(totals)
            if score > minimum and score not in scores:
                matched[f_index][t_index] = sum(totals)

    def get_best(f_index, t_indexes):
        if not t_indexes:
            return
        items = list(t_indexes.items())
        best_match = max(items, key=itemgetter(1))
        t_i = best_match[0]

        while t_i in assigned:
            try:
                old_match = matched[assigned[t_i]][t_i]
            except KeyError:
                break
            if best_match[1] > old_match:
                old_f_index = assigned[t_i]
                assigned[t_i] = f_index
                get_best(old_f_index, matched[old_f_index])
                return
            else:
                items.remove(best_match)
                if not items:
                    return
                best_match = max(items, key=itemgetter(1))
                t_i = best_match[0]

        assigned[t_i] = f_index

    for f_index, t_indexes in matched.items():
        best_match = max(list(t_indexes.items()), key=itemgetter(1))
        t_index = best_match[0]
        while t_index in assigned:
            prev_matched = assigned[t_index]
            if t_indexes[t_index] > matched[prev_matched][t_index]:
                get_best(prev_matched, matched[prev_matched])
                break
            else:
                del (t_indexes[t_index])
                if not t_indexes:
                    break
                best_match = max(list(t_indexes.items()), key=itemgetter(1))
                t_index = best_match[0]

        if t_indexes:
            assigned[t_index] = f_index

    ret_indexes = {}
    for t_index, f_index in assigned.items():
        try:
            ret[files[f_index].cls] = tracks[t_index]
            ret_indexes[t_index] = files[f_index].cls
        except AttributeError:
            ret[files[f_index]] = tracks[t_index]
            ret_indexes[t_index] = files[f_index]

    for t in tracks:
        if '#exact' in t:
            ret[t['#exact'].cls] = t
            del (t['#exact'])

    if jfdi:
        unmatched_tracks = [t for i, t in enumerate(tracks) if i
                            not in assigned]
        unmatched_files = [f.cls for f in files if f.cls not in ret]
        ret.update(brute_force_results(unmatched_files, unmatched_tracks))

    if existing:
        ret = dict((f, dict_difference(f, r)) for f, r in ret.items())

    if as_index:
        return ret, ret_indexes
    return ret


def merge_track(audio, info):
    track = {}

    for key in info.keys():
        if not key.startswith('#'):
            if isinstance(info[key], str):
                track[key] = info[key]
            else:
                if isinstance(info[key], list):
                    track[key] = info[key][::]
                elif isinstance(info[key], dict):
                    track[key] = info[key]

    for key in audio.keys():
        if not key.startswith('#'):
            if isinstance(audio[key], str):
                track[key] = audio[key]
            else:
                track[key] = audio[key][::]
    if '#exact' in audio:
        track['#exact'] = audio['#exact']
    return track


def merge_tsp_tracks(profiles, files=None):
    ret = []
    to_repl = []
    for tsp in profiles:
        if not tsp.matched:
            continue

        if tsp.result.tracks is None and files is not None:
            info = strip_fields(tsp.result.info, tsp.fields, leave_exact=True)
            tags = [deepcopy(info) for z in files]
        else:
            tags = [strip_fields(t, tsp.fields, leave_exact=True)
                    for t in tsp.result.merged]

        if len(tags) > len(ret):
            ret.extend(tags[len(ret):])
        if tsp.replace_fields:
            to_repl.append([tags, tsp.replace_fields])
        for i, t in enumerate(tags):
            ret[i] = combine_tracks(ret[i], t)

    for tracks, fields in to_repl:
        for repl, track in zip(tracks, ret):
            track.update(strip_fields(repl, fields, leave_exact=True))

    return ret


def masstag(mtp, files=None, flag=None, mtp_error_func=None,
            tsp_error_func=None, print_status=True):
    not_found = []
    found = []

    if files is None:
        files = mtp.files

    if flag is None:
        flag = MassTagFlag()
    elif flag.stop:
        return []

    assert files

    artist, album = get_artist_album(files)

    if artist and album:
        set_status(SEARCHING_ARTIST_ALBUM.format(artist, album))
    elif artist:
        set_status(SEARCHING_ARTIST.format(artist))
    elif album:
        set_status(SEARCHING_ALBUM.format(album))
    else:
        set_status(SEARCHING_NO_INFO)

    mtp.regexps = DEFAULT_REGEXP if not mtp.regexps else mtp.regexps

    for matches, results, tsp in mtp.search(files, errors=mtp_error_func):
        if flag.stop:
            break
        if len(results) > 1:
            set_status(RESULTS_FOUND.format(len(results)))
        elif not results:
            set_status(NO_RESULTS_FOUND)
        else:
            set_status(ONE_RESULT_FOUND)

        if not matches:
            not_found.append(tsp)
            continue

        if len(matches) > 1:
            set_status(MATCHING_ALBUMS_FOUND.format(len(matches)))
        else:
            set_status(ONE_MATCHING_ALBUM_FOUND)

        set_status(get_match_str(matches[0].info))
        result = tsp.retrieve(matches[0], errors=tsp_error_func)

        while not check_result(result, files):
            del (matches[0])
            if matches:
                set_status(RETRIEVING_NEXT)
                set_status(get_match_str(matches[0].info))
                result = tsp.retrieve(matches[0], errors=tsp_error_func)
            else:
                result = None
                break

        if result is None:
            set_status(NO_MATCHES.format(tsp.tag_source.name))
            not_found.append(tsp)
        else:
            found.append(tsp)

    ret = []

    if not_found and found:
        set_status(RECHECKING.format(found[0].tag_source.name))
        audios_copy = []
        for t, m in zip(list(map(deepcopy, files)), found[0].result.merged):
            audios_copy.append(combine_tracks(t, m))

        new_mtp = MassTagProfile(translate("Masstagging", 'Rechecking'),
                                 files=audios_copy, profiles=not_found,
                                 album_bound=mtp.album_bound, track_bound=mtp.track_bound,
                                 regexps=mtp.regexps)

        ret = masstag(new_mtp, audios_copy, flag,
                      mtp_error_func, tsp_error_func, False)

    if found:
        if not ret and print_status:
            set_status(VALID_FOUND)
        return [tsp.result for tsp in found] + ret
    else:
        if print_status:
            set_status(NO_VALID_FOUND)
        return []


def replace_tracknumbers(files, tracks):
    if len(files) != len(tracks):
        return

    files = sorted(files, key=lambda f: to_string(f.get('track', f[FILENAME])))
    try:
        tracks = sorted(tracks, key=itemgetter('track'))
        tracks = sorted(tracks, key=itemgetter('discnumber'))
    except KeyError:
        return

    if len(files) == len(tracks):
        discnum = 1
        track_count = 0
        offset = 0
        for f, t in zip(files, tracks):
            track_count += 1
            try:
                new_discnum = to_int(t['discnumber'])
            except (ValueError, TypeError):
                continue
            if new_discnum > discnum:
                offset = track_count - 1
                discnum = new_discnum
            try:
                f_tracknum = to_int(f['track'])
                t_tracknum = to_int(t['track'])
            except (ValueError, TypeError, KeyError):
                continue
            if f_tracknum > t_tracknum:
                f['track'] = [str(f_tracknum - offset)]


def split_files(audios, pattern):

    def copy_audio(f):
        tags = filenametotag(pattern, f['__path'], True)
        audio_copy = deepcopy(f)
        audio_copy.update(dict_difference(audio_copy, tags))
        audio_copy.cls = f
        return audio_copy

    tag_groups = []

    for dirpath, files in sorted_split_by_field(audios, '__dirpath'):
        album_groups = sorted_split_by_field(files, 'album')
        for album, album_files in album_groups:
            tag_groups.append(list(map(copy_audio, album_files)))

    return tag_groups


def to_int(v):
    return int(to_string(v))


class MassTagProfile(object):

    def __init__(self, name=DEFAULT_NAME, desc='', fields=None, files=None,
                 file_pattern=DEFAULT_PATTERN, profiles=None, album_bound=0.50,
                 track_bound=0.80, jfdi=True, leave_existing=False, regexps=''):

        object.__init__(self)

        self.album_bound = album_bound
        self.desc = desc
        self.fields = ['artist', 'title'] if fields is None else fields
        self.file_pattern = file_pattern
        self.files = [] if files is None else files
        self.jfdi = jfdi
        self.leave_existing = leave_existing
        self.name = name
        self.profiles = profiles if profiles is not None else []
        self.regexps = regexps
        self.track_bound = track_bound

    def clear(self):
        for profile in self.profiles:
            profile.clear_results()

    def search(self, files=None, profiles=None, regexps=None, errors=None):
        files = self.files if files is None else files
        profiles = self.profiles if profiles is None else profiles
        regexps = self.regexps if regexps is None else regexps

        if not files or not profiles:
            return

        self.files = files

        if regexps:
            changed_files = \
                [apply_regexps(f, regexps) for f in files]
            rxp_album = changed_files[0][0]
            changed_files = [z[1] for z in changed_files]

        for profile in profiles:
            profile.clear_results()
            set_status(POLLING.format(profile.tag_source.name))
            try:
                results = profile.search(files)
                if regexps and rxp_album:
                    profile.clear_results()
                    set_status(translate('Masstagging',
                                         "Retrying search with album name: <b>{}</b>"
                                         ).format(rxp_album))
                    rxp_results = profile.search(changed_files)
                    results.extend(rxp_results)
                    profile.clear_results()
                profile.results = results
            except RetrievalError as e:
                if errors is None:
                    raise e
                if errors(e, profile):
                    raise e
                yield [], [], profile
                continue

            if results:
                profile.find_matches(self.album_bound, files)
            profile.files = files
            yield profile.matched, profile.results, profile


class Result(object):

    def __init__(self, info=None, tracks=None, tag_source=None):
        object.__init__(self)

        self.__info = {}
        self.__tracks = []
        self.merged = []
        self.tag_source = tag_source

        self.tracks = tracks if tracks is not None else []
        self.info = {} if info is None else info
        self.track_matched = {}

    def _update_merged(self):
        if self.__tracks:
            self.merged = [merge_track(a, self.__info) for a in self.__tracks]
        else:
            self.merged = []

    @property
    def info(self):
        return self.__info

    @info.setter
    def info(self, value):
        self.__info = value
        self._update_merged()

    @property
    def tracks(self):
        return self.__tracks

    @tracks.setter
    def tracks(self, value):
        self.__tracks = value
        self._update_merged()

    def retrieve(self, errors=None):
        if self.tag_source:
            try:
                self.info, self.tracks = self.tag_source.retrieve(self.info)
            except RetrievalError as e:
                if errors is None:
                    raise
                if errors(e):
                    raise e
                else:
                    return {}, []
            return self.info, self.tracks
        return {}, []


class TagSourceProfile(object):

    def __init__(self, files=None, tag_source=None, fields=None,
                 if_no_result=CONTINUE, replace_fields=None):

        object.__init__(self)

        self.if_no_result = if_no_result
        self.fields = [] if fields is None else fields
        self.files = [] if files is None else files
        self.group_by = tag_source.group_by if tag_source else None
        self.matched = []
        self.replace_fields = [] if replace_fields is None else replace_fields
        self.result = None
        self.results = []
        self.tag_source = tag_source

    def clear_results(self):
        self.result = None
        self.results = []
        self.matched = []

    def find_matches(self, album_bound, files=None, results=None):
        files = self.files if files is None else files
        results = self.results if results is None else results

        assert files
        assert results

        self.matched = find_best(results, files, album_bound)
        return self.matched

    def retrieve(self, result, errors=None):
        info = result.info if hasattr(result, 'info') else result

        try:
            index = self.results.index(result)
        except ValueError:
            index = None

        try:
            results = self.tag_source.retrieve(info)
            if results is not None:
                self.result = Result(*results)
            else:
                self.result = result
                return self.result
        except RetrievalError as e:
            if errors is None:
                raise
            if errors(e, self):
                raise e
            else:
                self.result = Result({}, [])
        self.result.tag_source = self.tag_source
        if index is not None:
            self.results[index] = self.result
        return self.result

    def search(self, files=None, tag_source=None):
        tag_source = self.tag_source if tag_source is None else tag_source
        files = self.files if files is None else files

        assert hasattr(tag_source, 'search')
        assert files

        files = split_by_field(files, *tag_source.group_by)
        search_value = list(files.keys())[0]
        self.results = [Result(*x) for x in tag_source.search(search_value, files[search_value])]
        for r in self.results:
            r.tag_source = self.tag_source
        return self.results


if __name__ == '__main__':
    from .. import puddletag

    puddletag.load_plugins()
    from ..tagsources import tagsources

    sources = dict((t.name, t) for t in tagsources)
    source = sources['Local TSource Plugin']()
    source.applyPrefs(['/mnt/multimedia/testlib'])
    print(source._dirs)



# -*- coding: utf-8 -*-
import logging
import re

from pyparsing import (CaselessLiteral, Combine, OpAssoc, ParserElement,
                       QuotedString, Word, alphanums, infix_notation)


from . import findfunc, audioinfo
from .puddleobjects import gettaglist
from .util import to_string


ParserElement.enable_packrat()


def str_cmp(a, b):
    if not isinstance(a, str):
        a = '\\'.join(a)

    if not isinstance(b, str):
        b = '\\'.join(b)

    return a.lower() == b.lower()


FIELDS = set(z.lower() for z in gettaglist()).union(audioinfo.FILETAGS)


def parse_arg(audio, text):
    if not isinstance(text, str):
        return text
    if text[0] == '%' and text[-1] == '%':
        return to_string(audio.get(text[1:-1], ''))
    elif text in FIELDS:
        return to_string(audio.get(text, ''))
    else:
        if text[0] == '"' and text[-1] == '"':
            text = text[1:-1]
        return findfunc.parsefunc(text, audio)


def wrap_bool(original):
    def __bool__(self):
        if hasattr(self, 'args'):
            self.args = [parse_arg(self.audio, z) for z in self.args]
        else:
            self.arg = parse_arg(self.audio, self.arg)
        return original(self)

    return __bool__


class BoolOperand(object):
    def __init__(self, t):
        self.args = t[0][0::2]


class BoolAnd(BoolOperand):
    @wrap_bool
    def __bool__(self):
        logging.debug('and: ' + str(self.args))
        for a in self.args:
            if not bool(a):
                return False
        return True


class BoolOr(BoolOperand):
    @wrap_bool
    def __bool__(self):
        logging.debug('or: ' + str(self.args))
        for a in self.args:
            if bool(a):
                return True
        return False


class BoolNot(BoolOperand):
    def __init__(self, t):
        self.arg = t[0][1]

    @wrap_bool
    def __bool__(self):
        logging.debug('not: ' + str(self.arg))
        if isinstance(self.arg, str):
            arg = self.arg.lower()
            for v in self.audio.values():
                if isinstance(v, str):
                    v = [v]
                v = '\\\\'.join(v).lower()
                if arg in v:
                    return False
            return True
        return not bool(self.arg)


class Greater(BoolOperand):
    @wrap_bool
    def __bool__(self):
        logging.debug('greater: ' + str(self.args))
        try:
            self.args = list(map(float, self.args))
        except ValueError:
            pass
        return self.args[0] > self.args[1]


class Less(BoolOperand):
    @wrap_bool
    def __bool__(self):
        logging.debug('less: ' + str(self.args))
        try:
            self.args = list(map(float, self.args))
        except ValueError:
            pass
        return self.args[0] < self.args[1]


class Equal(BoolOperand):
    @wrap_bool
    def __bool__(self):
        logging.debug('equal: ' + str(self.args))
        return str_cmp(self.args[0], self.args[1])


class Missing(BoolOperand):
    def __init__(self, t):
        self.arg = t[0][1]

    def __bool__(self):
        logging.debug('missing: ' + str(self.arg))
        if getattr(self, "audio", None):
            return not (self.arg in self.audio)
        return False


class Present(BoolOperand):
    def __init__(self, t):
        self.arg = t[0][1]

    def __bool__(self):
        logging.debug('present: ' + str(self.arg))
        if getattr(self, "audio", None):
            return (self.arg in self.audio)
        return False


class BoolIs(BoolOperand):
    @wrap_bool
    def __bool__(self):
        logging.debug('is: ' + str(self.args))
        return str_cmp(self.args[0], self.args[1])


class Has(BoolOperand):
    @wrap_bool
    def __bool__(self):
        logging.debug('has: ' + str(self.args))
        return self.args[1].lower() in self.args[0].lower()


class Matches(BoolOperand):
    @wrap_bool
    def __bool__(self):
        logging.debug('matches: ' + str(self.args))
        return not re.search(self.args[1].lower(), self.args[0].lower()) is None


bool_exprs = [
    (CaselessLiteral("missing"), 1, OpAssoc.RIGHT, Missing),
    (CaselessLiteral("present"), 1, OpAssoc.RIGHT, Present),
    (CaselessLiteral("greater"), 2, OpAssoc.LEFT,  Greater),
    (CaselessLiteral("less"),    2, OpAssoc.LEFT,  Less),
    (CaselessLiteral("equal"),   2, OpAssoc.LEFT,  Equal),
    (CaselessLiteral("has"),     2, OpAssoc.LEFT,  Has),
    (CaselessLiteral("matches"), 2, OpAssoc.LEFT,  Matches),
    (CaselessLiteral("is"),      2, OpAssoc.LEFT,  BoolIs),
    (CaselessLiteral("and"),     2, OpAssoc.LEFT,  BoolAnd),
    (CaselessLiteral("or"),      2, OpAssoc.LEFT,  BoolOr),
    (CaselessLiteral("not"),     1, OpAssoc.RIGHT, BoolNot),
]

field_expr = Combine('%' + Word(alphanums + '_') + '%')
tokens = QuotedString('"', unquote_results=False) \
         | field_expr | Word(alphanums + '_')
bool_expr = infix_notation(tokens, bool_exprs)


def parse(audio, expr):
    for i in bool_exprs:
        i[3].audio = audio
    try:
        res = bool_expr.parse_string(expr)[0]
    except ParseException as e:
        res = expr
    if isinstance(res, str):
        res = res.lower()
        for field, value in audio.items():
            if isinstance(value, str):
                value = [value]
            elif isinstance(value, (int, float)):
                value = [str(value)]
            try:
                logging.debug('simple filter: %s in %s', res, value)
                if res in '\\\\'.join(value).lower():
                    return True
            except TypeError as e:
                continue
    else:
        return bool(res)
    return False


if __name__ == '__main__':
    audio = audioinfo.Tag('clen.mp3')
    # parse(audio, "not p")
    # parse(audio, 'not missing artist')
    # parse(audio, '7 greater 6')
    # parse(audio, '%track% greater 14')
    # parse(audio, '%track% greater "$add($len(%artist%), 50)"')
    # t = time.time()
    # parse(audio, '(not missing artist) and (20 greater 19)')
    # parse(audio, 'not (20 greater 19)')
    # print time.time() - t
    # parse(audio, 'not missing artist and 18 greater 19')
    # parse(audio, 'artist is "Carl Douglas"')
    # parse(audio, "artist has aarl")
    # parse(audio, "artist has Carl")
    import time

    t = time.time()
    print(audio.filepath)
    print(parse(audio, '__filename has clen'))



from .constants import *
from .util import *

AbstractTag = MockTag

extensions = {}
options = []

mapping = {}
revmapping = {}


def loadmapping(filepath, default=None):
    try:
        lines = open(filepath, 'r').read().split('\n')
    except (IOError, OSError):
        if default:
            return default
        else:
            return {}
    mappings = {}
    for l in lines:
        tags = [z.strip() for z in l.split(',')]
        if len(tags) == 3:  # Tag, Source, Target
            try:
                mappings[tags[0]].update({tags[1]: tags[2]})
            except KeyError:
                mappings[tags[0]] = ({tags[1]: tags[2]})
    return mappings


def register_tag(mut_obj, tag, tag_name, tag_exts=None):
    if isinstance(tag_exts, str):
        extensions[tag_exts] = [mut_obj, tag, tag_name]
    elif tag_exts is not None:
        for e in tag_exts:
            extensions[e] = [mut_obj, tag, tag_name]

    options.append([mut_obj, tag, tag_name])


def setmapping(m):
    global revmapping
    global mapping

    mapping = m
    for z in mapping:
        revmapping[z] = CaselessDict([(value, key) for key, value in mapping[z].items()])
    for z in extensions.values():
        try:
            if z[2] in mapping:
                z[1].mapping = mapping[z[2]]
                z[1].revmapping = revmapping[z[2]]
            if 'global' in mapping:
                z[1].mapping.update(mapping['global'])
                z[1].revmapping.update(revmapping['global'])
        except IndexError:
            pass


def Tag(filename):
    """Class that operates on audio tags.
    Currently supports ogg, mp3, mp4, apev2 and flac files

    It can be used in two ways.

    >>>tag = audioinfo.Tag(filename)
    Gets the tags in the audio, filename
    as a dictionary in format {tag: value} in Tag._tags.

    On the other hand, if you have already created
    a tag object. Use link like so:

    >>>tag = audioinfo.Tag()
    >>>tag.link(filename)
    {'artist': "Artist", "track":"12", title:"Title", '__length':"5:14"}

    File info tags like length start with '__'.
    Images can be accessed by either the '__image' tag or via Tag.images. Note
    that images aren't included when iterating through Tag.

    Use save to save tags.

    There are caveats associated with each module, so check out their docstrings
    for more info."""

    fileobj = open(filename, "rb")
    ext = splitext(filename)
    try:
        return extensions[ext][1](filename)
    except KeyError:
        pass

    try:
        header = fileobj.read(128)
        results = [Kind[0].score(filename, fileobj, header) for Kind in options]
    finally:
        fileobj.close()
    results = list(zip(results, options))
    results.sort(key=lambda v: v[0])
    score, Kind = results[-1]
    if score > 0:
        return Kind[1](filename)
    else:
        return None


from . import id3, vorbis, apev2, mp4, wma

tag_modules = (id3, vorbis, apev2, mp4, wma)

for m in tag_modules:
    if hasattr(m, 'filetype'):
        register_tag(*m.filetype)
    if hasattr(m, 'filetypes'):
        list(map(lambda x: register_tag(*x), m.filetypes))

setmapping({'VorbisComment': {'tracknumber': 'track', 'date': 'year'}})

_Tag = Tag
model_tag = lambda x: x



# -*- coding: utf-8 -*-
import os

from PyQt5.QtWidgets import QApplication, QWidget, QCheckBox, QVBoxLayout

from .constants import CONFIGDIR
from .puddleobjects import PuddleConfig
from .translations import translate

NAME = 'name'
DESC = 'description'
SECTION = 'Config'
VALUE = 'value'

_filename = os.path.join(CONFIGDIR, 'confirmations')
_confirmations = {}
_registered = []


def add(name, default=True, desc=None):
    if desc is None:
        desc = name
    if name not in _confirmations:
        _confirmations[name] = [default, desc]
        if name not in _registered:
            _registered.append(name)


def should_show(name):
    return _confirmations[name][0]


def _load(filename):
    cparser = PuddleConfig(filename)
    confirmations = {}
    for section in cparser.sections():
        if section.startswith(SECTION):
            name = cparser.get(section, NAME, '')
            desc = cparser.get(section, DESC, '')
            value = cparser.get(section, VALUE, True)
            confirmations[name] = [value, desc]
    return confirmations


def load():
    # _confirmations.clear()
    _confirmations.update(_load(_filename))


def save(filename=None, confirmations=None):
    if filename is None:
        filename = _filename
    cparser = PuddleConfig(filename)

    if confirmations is None:
        confirmations = _confirmations

    for i, name in enumerate(confirmations):
        set_value = lambda k, v: cparser.set(SECTION + str(i), k, v)
        set_value(NAME, name)
        set_value(VALUE, confirmations[name][0])
        set_value(DESC, confirmations[name][1])


class Settings(QWidget):
    def __init__(self, parent=None):
        super(Settings, self).__init__(parent)
        layout = QVBoxLayout()
        load()
        self._controls = {}
        for name in _registered:
            control = QCheckBox(translate('Confirmations',
                                          _confirmations[name][1]))
            control.setChecked(_confirmations[name][0])
            layout.addWidget(control)
            self._controls[name] = control
        layout.addStretch()
        self.setLayout(layout)

    def applySettings(self, control=None):
        for name, control in self._controls.items():
            _confirmations[name][0] = control.isChecked()
        save()


if __name__ == '__main__':
    app = QApplication([])
    add('First True', True)
    add('Name', False, 'Description')
    print(_confirmations)
    win = Settings()
    win.show()
    app.exec_()



import contextlib
import http.client
import os
import time
import traceback
import urllib.error
import urllib.parse
import urllib.request
from itertools import chain, product, starmap

import acoustid

from puddlestuff import audioinfo

from puddlestuff.audioinfo import stringtags
from puddlestuff.constants import SPINBOX, TEXT
from puddlestuff.tagsources import set_status, write_log, SubmissionError
from puddlestuff.tagsources.musicbrainz import retrieve_album
from puddlestuff.translations import translate
from puddlestuff.util import escape_html, isempty, to_string

CALCULATE_MSG = translate('AcoustID', "Calculating ID")
RETRIEVE_MSG = translate('AcoustID', "Retrieving AcoustID data: {} of {}.")
RETRIEVE_MB_MSG = translate('AcoustID', "Retrieving MB album data: {}")
FP_ERROR_MSG = translate('AcoustID', "Error generating fingerprint: {}")
WEB_ERROR_MSG = translate('AcoustID', "Error retrieving data: {}")
SUBMIT_ERROR_MSG = translate('AcoustID', "Error submitting data: {}")
SUBMIT_MSG = translate('AcoustID', "Submitting data to AcoustID: {} to {} of {}.")
FOUND_ID_MSG = translate('AcoustID', "Found AcoustID in file.")
FILE_MSG = translate('AcoustID', "File #{}: {}")

API_KEY = "gT8GJxhO"


def album_hash(d):
    h = ''
    if 'album' in d:
        h = d['album']

    if 'year' in d:
        h += d['year']

    return hash(h)


def best_match(albums, tracks):
    hashed = {}
    data = (product(a, [t]) for a, t in zip(albums, tracks))

    for album, track in chain(*data):
        key = album_hash(album)
        if key not in hashed:
            hashed[key] = [album, 1, [track]]
        else:
            hashed[key][1] += 1
            hashed[key][2].append(track)

    matched_tracks = []
    ret = []

    for key in sorted(hashed, key=lambda i: hashed[i][1], reverse=True):
        album, count, tracks = hashed[key]
        new_tracks = []

        for t in tracks:
            if t['#exact'] not in matched_tracks:
                new_tracks.append(t)
                matched_tracks.append(t['#exact'])
        if not new_tracks: continue
        ret.append([album, new_tracks])

    return ret


def convert_for_submit(tags):
    cipher = {
        'mbrainz_track_id': 'mbid',
        'title': 'track',
        'track': 'trackno',
        'discnumber': 'discno',
        '__bitrate': 'bitrate',
        'musicip_puid': 'puid',
    }

    valid_keys = set(['artist', 'album', 'title', 'track', 'discno',
                      'mbid', 'year', 'bitrate', 'puid', 'trackno'])

    ret = dict((cipher.get(k, k), v) for k, v in stringtags(tags).items()
               if cipher.get(k, k) in valid_keys and v)
    bitrate = ret['bitrate'].split(' ')[0]
    if bitrate == 0:
        del (ret['bitrate'])
    else:
        ret['bitrate'] = str(bitrate)

    return ret


def fingerprint_file(fn):
    return acoustid._fingerprint_file_fpcalc(fn, 120)


def id_in_tag(tag):
    if 'acoustid_fingerprint' in tag:
        fp = to_string(tag['acoustid_fingerprint'])
    else:
        return

    if '__length' in tag:
        duration = audioinfo.lnglength(tag['__length'])
    else:
        return

    return (duration, fp)


def match(apikey, path, fp=None, dur=None, meta='releases recordings tracks'):
    """Look up the metadata for an audio file. If ``parse`` is true,
    then ``parse_lookup_result`` is used to return an iterator over
    small tuple of relevant information; otherwise, the full parsed JSON
    response is returned.
    """
    path = os.path.abspath(os.path.expanduser(path))
    if None in (fp, dur):
        dur, fp = fingerprint_file(path)
    response = acoustid.lookup(apikey, fp, dur, meta)
    return response, fp


def parse_release_data(rel):
    info = {}
    info['__numtracks'] = str(rel.get('track_count', ''))
    info['album'] = rel.get('title', '')

    if 'date' in rel:
        date = rel['date']
        info['year'] = '-'.join(str(z).zfill(2) for z in
                                map(date.get, ('year', 'month', 'day')) if z)
    info['country'] = rel.get('country', '')
    info['discs'] = str(rel.get('medium_count', ''))
    info['#album_id'] = rel['id']
    info['mbrainz_album_id'] = rel['id']
    if 'mediums' in rel:
        info['track'] = str(
            rel['mediums'][0]['tracks'][0].get('position', ""))
    return dict((k, v) for k, v in info.items() if not isempty(v))


def parse_lookup_result(data, albums=False, fp=None):
    if data['status'] != 'ok':
        raise acoustid.WebServiceError("status: %s" % data['status'])
    if 'results' not in data:
        raise acoustid.WebServiceError("results not included")

    try:
        result = data['results'][0]
    except IndexError:
        return None
    info = {}
    info['#score'] = result['score']
    if not result.get('recordings'):
        # No recording attached. This result is not very useful.
        return {
            'acoustid_id': result['id'],
            '#score': result['score'],
            'acoustid_fingerprint': fp, }

    if fp:
        info['acoustid_fingerprint'] = fp

    tracks = [parse_recording_data(r, info) for r in result['recordings']]

    return tracks


def parse_recording_data(data, info=None):
    track = {} if info is None else info.copy()

    try:
        track['title'] = data['title']
    except KeyError:
        track['acoustid_id'] = data['id']
        return {}, track
    if 'duration' in data:
        track['__length'] = audioinfo.strlength(data['duration'])
    track['acoustid_id'] = data['id']

    track['artist'] = data.get('artists', [{'name': ""}])[0]['name']
    if track['artist']:
        track['mbrainz_artist_id'] = data['artists'][0]['id']

    if 'releases' in data:
        album_info = list(map(parse_release_data, data['releases']))
    else:
        album_info = []

    track = dict((k, v) for k, v in track.items() if not isempty(v))

    if 'artist' in track:
        for album in album_info:
            if 'artist' not in album:
                album['artist'] = track['artist']

    return album_info, track


def retrieve_album_info(album, tracks):
    if not album:
        return album, tracks
    msg = '<b>%s - %s</b>' % tuple(map(escape_html,
                                       (album['artist'], album['album'])))
    msg = RETRIEVE_MB_MSG.format(msg)
    write_log(msg)
    set_status(msg)

    info, new_tracks = retrieve_album(album['mbrainz_album_id'])
    for t in tracks:
        try:
            index = int(t['track'])
        except KeyError:
            for index, nt in enumerate(new_tracks):
                if nt['title'] == t['title']:
                    break
        t.update(new_tracks[index])
        new_tracks[index] = t

    return info, new_tracks


def which(program):

    # http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


class AcoustID(object):
    name = 'AcoustID'
    group_by = ['album', None]

    def __init__(self):
        object.__init__(self)
        self.min_score = 0.80
        self.preferences = [
            [translate("AcoustID", 'Minimum Score'), SPINBOX, [0, 100, 80]],
            [translate("AcoustID", "AcoustID Key"), TEXT, ""]
        ]
        self.__lasttime = time.time()
        acoustid._send_request = self._send_request
        self.__user_key = ""

    def _send_request(self, req):
        """Given a urllib2 Request object, make the request and return a
        tuple containing the response data and headers.
        """
        if time.time() - self.__lasttime < 0.4:
            time.sleep(time.time() - self.__lasttime + 0.1)
        try:
            with contextlib.closing(urllib.request.urlopen(req)) as f:
                return f.read(), f.info()
        except urllib.error.HTTPError as exc:
            raise acoustid.WebServiceError('HTTP status %i' % exc.code, exc.read())
        except http.client.BadStatusLine:
            raise acoustid.WebServiceError('bad HTTP status line')
        except IOError:
            raise acoustid.WebServiceError('connection failed')

    def search(self, artist, fns=None):

        tracks = []
        albums = []

        fns_len = len(fns)
        for i, fn in enumerate(fns):
            try:
                disp_fn = audioinfo.decode_fn(fn.filepath)
            except AttributeError:
                disp_fn = fn['__path']
            write_log(disp_fn)
            try:

                fp = id_in_tag(fn)
                if fp:
                    write_log(FOUND_ID_MSG)
                    dur, fp = fp
                else:
                    write_log(CALCULATE_MSG)
                    dur, fp = (None, None)

                write_log(RETRIEVE_MSG.format(i + 1, fns_len))
                set_status(RETRIEVE_MSG.format(i + 1, fns_len))

                data, fp = match("gT8GJxhO", fn.filepath, fp, dur)
                write_log(translate('AcoustID', "Parsing Data"))

                info = parse_lookup_result(data, fp=fp)
            except acoustid.FingerprintGenerationError as e:
                write_log(FP_ERROR_MSG.format(str(e)))
                continue
            except acoustid.WebServiceError as e:
                set_status(WEB_ERROR_MSG.format(str(e)))
                write_log(WEB_ERROR_MSG.format(str(e)))
                break

            if hasattr(info, 'items'):
                albums.append([{}])
                info['#exact'] = fn
                tracks.append(info)
            elif info is not None:
                for album, track in info:
                    if track and track['#score'] >= self.min_score:
                        track['#exact'] = fn
                        tracks.append(track)
                        albums.append(album if album else [{}])

        return starmap(retrieve_album_info, best_match(albums, tracks))

    def submit(self, fns):
        if not self.__user_key:
            raise SubmissionError(translate("AcoustID",
                                            "Please enter AcoustID user key in settings."))

        fns_len = len(fns)
        data = []
        for i, fn in enumerate(fns):

            try:
                disp_fn = audioinfo.decode_fn(fn.filepath)
            except AttributeError:
                disp_fn = fn['__path']
            write_log(FILE_MSG.format(i + 1, disp_fn))

            try:
                fp = id_in_tag(fn)
                if fp:
                    write_log(FOUND_ID_MSG)
                    dur, fp = fp
                else:
                    write_log(CALCULATE_MSG)
                    dur, fp = fingerprint_file(fn.filepath)

                info = {
                    'duration': str(dur),
                    'fingerprint': str(fp),
                }

                info.update(convert_for_submit(fn))
                data.append(info)

                if len(data) > 9 or i == fns_len - 1:
                    msg = SUBMIT_MSG.format(i - len(data) + 2, i + 1, fns_len)
                    write_log(msg)
                    set_status(msg)
                    acoustid.submit(API_KEY, self.__user_key, data)
                    data = []

            except acoustid.FingerprintGenerationError as e:
                traceback.print_exc()
                write_log(FP_ERROR_MSG.format(str(e)))
                continue
            except acoustid.WebServiceError as e:
                traceback.print_exc()
                set_status(SUBMIT_ERROR_MSG.format(str(e)))
                write_log(SUBMIT_ERROR_MSG.format(str(e)))
                break

    def retrieve(self, info):
        return None

    def applyPrefs(self, args):
        self.min_score = args[0] / 100.0
        self.__user_key = args[1]


if not which('fpcalc'):
    raise ImportError("fpcalc not found on system")

info = AcoustID

if __name__ == '__main__':
    x = AcoustID()
    x.applyPrefs([85, "KEIY0X4P"])
    file_dir = ''
    files = []
    for z in os.listdir(file_dir):
        fn = os.path.join(file_dir, z)
        try:
            tag = audioinfo.Tag(fn)
            if tag is not None:
                files.append(tag)
        except:
            pass
    x.submit(files)



import sys

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QAbstractItemView, QApplication, QComboBox, QFrame, QHBoxLayout, QInputDialog, QPushButton, \
    QShortcut, QVBoxLayout

from ..puddleobjects import (PuddleConfig, ListBox,
                             ListButtons)
from ..translations import translate


def load_patterns(filepath=None):
    settings = PuddleConfig(filepath)
    return [settings.get('editor', 'patterns',
                         ['%artist% - $num(%track%,2) - %title%',
                          '%artist% - %album% - $num(%track%,2) - %title%',
                          '%artist% - %title%', '%artist% - %album%',
                          '%artist% - Track %track%', '%artist% - %title%']),
            settings.get('editor', 'index', 0, True)]


class PatternCombo(QComboBox):
    name = 'toolbar-patterncombo'
    patternchanged = pyqtSignal(str, name='patternchanged')

    def __init__(self, items=None, parent=None, status=None):
        self.emits = ['patternchanged']
        self.receives = [('patterns', self.setItems)]
        self.settingsdialog = SettingsWin
        QComboBox.__init__(self, parent)
        self._status = status
        status['patterns'] = self.items
        status['patterntext'] = lambda: str(self.currentText())

        self.setEditable(True)
        if items:
            self.addItems(items)
        pchange = lambda text: self.patternchanged.emit(str(text))
        self.editTextChanged.connect(pchange)

        shortcut = QShortcut(self)
        shortcut.setKey('F8')
        shortcut.setContext(Qt.ShortcutContext.ApplicationShortcut)

        def set_focus():
            if self.hasFocus():
                status['table'].setFocus()
            else:
                self.lineEdit().selectAll()
                self.setFocus()

        shortcut.activated.connect(set_focus)

    def setItems(self, texts):
        self.clear()
        self.addItems(texts)

    def items(self):
        text = self.itemText
        return [str(text(i)) for i in range(self.count())]

    def loadSettings(self):
        patterns, index = load_patterns()
        self.setItems(patterns)
        self.setCurrentIndex(index)

    def saveSettings(self):
        settings = PuddleConfig()
        settings.set('editor', 'patterns', list(self.items()))
        settings.set('editor', 'index', self.currentIndex())


class SettingsWin(QFrame):
    def __init__(self, parent=None, cenwid=None, status=None):
        QFrame.__init__(self, parent)
        self.title = translate('Settings', "Patterns")
        connect = lambda c, signal, s: getattr(c, signal).connect(s)
        self.setFrameStyle(QFrame.Shape.Box)
        self.listbox = ListBox()
        self.listbox.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        buttons = ListButtons()

        self.listbox.addItems(status['patterns'])
        hbox = QHBoxLayout()
        hbox.addWidget(self.listbox)
        self.setLayout(hbox)

        vbox = QVBoxLayout()
        sortlistbox = QPushButton(translate("Pattern Settings", '&Sort'))
        self._sortOrder = Qt.SortOrder.AscendingOrder
        connect(sortlistbox, 'clicked', self._sortListBox)
        vbox.addWidget(sortlistbox)
        vbox.addLayout(buttons)
        vbox.addStretch()

        hbox.addLayout(vbox)

        connect(buttons, "add", self.addPattern)
        connect(buttons, "edit", self.editItem)
        buttons.duplicateButton.setVisible(False)
        self.listbox.connectToListButtons(buttons)
        self.listbox.editButton = buttons.editButton
        connect(self.listbox, 'itemDoubleClicked', self._doubleClicked)

    def _sortListBox(self):
        if self._sortOrder == Qt.SortOrder.AscendingOrder:
            self.listbox.sortItems(Qt.SortOrder.DescendingOrder)
            self._sortOrder = Qt.SortOrder.DescendingOrder
        else:
            self.listbox.sortItems(Qt.SortOrder.AscendingOrder)
            self._sortOrder = Qt.SortOrder.AscendingOrder

    def saveSettings(self):
        patterns = [str(self.listbox.item(row).text()) for row in range(self.listbox.count())]
        cparser = PuddleConfig()
        cparser.setSection('editor', 'patterns', patterns)

    def addPattern(self):
        l = self.listbox.item
        patterns = [str(l(z).text()) for z in range(self.listbox.count())]
        row = self.listbox.currentRow()
        if row < 0:
            row = 0
        (text, ok) = QInputDialog().getItem(self, 'puddletag',
                                            translate("Pattern Settings", 'Enter a pattern'), patterns, row)
        if ok:
            self.listbox.clearSelection()
            self.listbox.addItem(text)
            self.listbox.setCurrentRow(self.listbox.count() - 1)

    def _doubleClicked(self, item):
        self.editItem()

    def editItem(self, row=None):
        if row is None:
            row = self.listbox.currentRow()
        l = self.listbox.item
        patterns = [str(l(z).text()) for z in range(self.listbox.count())]
        (text, ok) = QInputDialog().getItem(self, 'puddletag',
                                            translate("Pattern Settings", 'Enter a pattern'),
                                            patterns, row)
        if ok:
            item = l(row)
            item.setText(text)
            item.setSelected(True)

    def applySettings(self, control):
        item = self.listbox.item
        patterns = [item(row).text() for row in range(self.listbox.count())]
        control.setItems(patterns)


control = ('patterncombo', PatternCombo, False)

if __name__ == '__main__':
    app = QApplication(sys.argv)
    win = PatternCombo(['one', 'the', 'three'])
    win.show()
    app.exec_()



from fnmatch import fnmatch
from os import listdir
from os.path import dirname

dirpath = dirname(__file__)
__all__ = [f[:-3] for f in listdir(dirpath) if fnmatch(f, '*.py') and
           f != '__init__.py']



# -*- coding: utf-8 -*-
"""Dialog's that crop up along the application, but are used at at most
one place, and aren't that complicated are put here."""
import os
import sys
from copy import deepcopy

from PyQt5.QtCore import QItemSelectionModel, Qt, pyqtRemoveInputHook, pyqtSignal
from PyQt5.QtGui import QPalette, QBrush, QColor
from PyQt5.QtWidgets import QAbstractItemView, QAction, QApplication, QCheckBox, QComboBox, QCompleter, \
    QDialog, QFileDialog, QFrame, QGridLayout, QHBoxLayout, QHeaderView, QLabel, QLineEdit, QMessageBox, \
    QPlainTextEdit, QPushButton, QSpinBox, QTableWidget, QTableWidgetItem, QTextEdit, QToolButton, QVBoxLayout

from . import findfunc, audioinfo
from .audioinfo import commontags, PATH
from .constants import HOMEDIR, KEEP
from .puddleobjects import (
    get_icon, gettaglist, partial,
    settaglist, winsettings, ListButtons, MoveButtons, OKCancel,
    PicWidget, PuddleConfig, natural_sort_key)
from .translations import translate
from .util import pprint_tag
from .util import to_string

ADD, EDIT, REMOVE = (1, 2, 3)
UNCHANGED = 0
BOLD = 1
ITALICS = 2
TAG_DISP = "<b>%s: </b> %s, "


class AutonumberDialog(QDialog):
    newtracks = pyqtSignal(int, int, 'Qt::CheckState', int, str, str, 'Qt::CheckState', name='newtracks')

    def __init__(self, parent=None, minval=0, numtracks=0,
                 enablenumtracks=False):

        QDialog.__init__(self, parent)

        self.setWindowTitle(
            translate('Autonumbering Wizard', "Autonumbering Wizard"))
        winsettings('autonumbering', self)

        def hbox(*widgets):
            box = QHBoxLayout()
            [box.addWidget(z) for z in widgets]
            box.addStretch()
            return box

        vbox = QVBoxLayout()

        self._start = QSpinBox()
        self._start.setValue(minval)
        self._start.setMaximum(65536)

        startlabel = QLabel(translate('Autonumbering Wizard', "&Start: "))
        startlabel.setBuddy(self._start)

        vbox.addLayout(hbox(startlabel, self._start))

        self._padlength = QSpinBox()
        self._padlength.setValue(1)
        self._padlength.setMaximum(65535)
        self._padlength.setMinimum(1)

        label = QLabel(translate('Autonumbering Wizard',
                                 'Max length after padding with zeroes: '))
        label.setBuddy(self._padlength)

        vbox.addLayout(hbox(label, self._padlength))

        self._separator = QCheckBox(translate('Autonumbering Wizard',
                                              "Add track &separator ['/']: Number of tracks"))
        self._numtracks = QSpinBox()
        self._numtracks.setEnabled(False)
        self._numtracks.setMaximum(65535)
        if numtracks:
            self._numtracks.setValue(numtracks)
        self._restart_numbering = QCheckBox(translate('Autonumbering Wizard',
                                                      "&Restart numbering at each directory group."))
        self._restart_numbering.stateChanged.connect(
            self.showDirectorySplittingOptions)

        vbox.addLayout(hbox(self._separator, self._numtracks))
        vbox.addWidget(self._restart_numbering)

        self.custom_numbering_widgets = []

        label = QLabel(translate('Autonumbering Wizard', "Group tracks using pattern:: "))

        self.grouping = QLineEdit()
        label.setBuddy(self.grouping)

        vbox.addLayout(hbox(label, self.grouping))
        self.custom_numbering_widgets.extend([label, self.grouping])

        label = QLabel(translate('Autonumbering Wizard', "Output field: "))

        self.output_field = QComboBox()
        label.setBuddy(self.output_field)

        self.output_field.setEditable(True)
        completer = self.output_field.completer()
        completer.setCaseSensitivity(Qt.CaseSensitivity.CaseSensitive)
        completer.setCompletionMode(QCompleter.CompletionMode.UnfilteredPopupCompletion)

        self.output_field.setCompleter(completer)
        self.output_field.addItems(gettaglist())
        vbox.addLayout(hbox(label, self.output_field))
        self.custom_numbering_widgets.extend([label, self.output_field])

        self.count_by_group = QCheckBox(translate('Autonumbering Wizard',
                                                  'Increase counter only on group change'))
        vbox.addWidget(self.count_by_group)
        self.custom_numbering_widgets.append(self.count_by_group)

        okcancel = OKCancel()
        vbox.addLayout(okcancel)
        self.setLayout(vbox)

        okcancel.ok.connect(self.emitValuesAndSave)
        okcancel.cancel.connect(self.close)
        self._separator.stateChanged.connect(
            lambda v: self._numtracks.setEnabled(v == Qt.CheckState.Checked))

        # self._restart_numbering.stateChanged.connect(
        #              self.showDirectorySplittingOptions)

        self._separator.setChecked(enablenumtracks)

        self._loadSettings()

    def showDirectorySplittingOptions(self, state):
        is_checked = state == Qt.CheckState.Checked
        for widget in self.custom_numbering_widgets:
            widget.setVisible(is_checked)

        if is_checked:
            self._numtracks.setVisible(False)
            self._separator.setText(translate('Autonumbering Wizard',
                                              "Add track &separator ['/']"))
        else:
            self._numtracks.setVisible(True)
            self._separator.setText(translate('Autonumbering Wizard',
                                              "Add track &separator ['/']: Number of tracks"))

    def emitValuesAndSave(self):
        if self._numtracks.isVisible():
            if self._separator.isChecked():
                numtracks = self._numtracks.value()
            else:
                numtracks = -1  # Don't use totals
        else:
            if self._separator.isChecked():
                numtracks = -2  # Use totals, automaticall generated
            else:
                numtracks = -1

        self.close()

        self.newtracks.emit(
            self._start.value(),
            numtracks,
            self._restart_numbering.checkState(),
            self._padlength.value(),
            str(self.grouping.text()),
            str(self.output_field.currentText()),
            self.count_by_group.checkState()
        )

        self._saveSettings()

    def _loadSettings(self):
        cparser = PuddleConfig()
        section = 'autonumbering'
        self._start.setValue(cparser.get(section, 'start', 1))
        self._separator.setCheckState(
            cparser.get(section, 'separator', Qt.CheckState.Unchecked))
        self._padlength.setValue(cparser.get(section, 'padlength', 1))

        self._restart_numbering.setCheckState(
            cparser.get(section, 'restart', Qt.CheckState.Unchecked))

        self.count_by_group.setCheckState(
            cparser.get(section, 'count_by_group', Qt.CheckState.Unchecked))

        self.showDirectorySplittingOptions(self._restart_numbering.checkState())

        self.grouping.setText(cparser.get(section, 'grouping', '%__dirpath%'))

        output_field_text = cparser.get(section, 'output_field', 'track')
        if not output_field_text:
            output_field_text = 'track'

        last_output_field_index = self.output_field.findText(output_field_text)
        if last_output_field_index > -1:
            self.output_field.setCurrentIndex(last_output_field_index)

    def _saveSettings(self):
        cparser = PuddleConfig()
        section = 'autonumbering'
        cparser.set(section, 'start', self._start.value())
        cparser.set(section, 'separator', self._separator.checkState())
        cparser.set(section, 'count_by_group', self.count_by_group.checkState())
        cparser.set(section, 'numtracks', self._numtracks.value())
        cparser.set(section, 'restart', self._restart_numbering.checkState())
        cparser.set(section, 'padlength', self._padlength.value())
        cparser.set(section, 'grouping', self.grouping.text())
        cparser.set(section, 'output_field', self.output_field.currentText())


class ImportTextFile(QDialog):
    """Dialog that importing a text file to retrieve tags from."""
    Newtags = pyqtSignal(list, str, name='Newtags')

    def __init__(self, parent=None, filename=None, clipboard=None):
        QDialog.__init__(self, parent)

        self.setWindowTitle(
            translate('Text File -> Tag', "Import tags from text file"))
        winsettings('importwin', self)

        grid = QGridLayout()

        self.label = QLabel(translate('Text File -> Tag', "Text"))
        grid.addWidget(self.label, 0, 0)

        self.label = QLabel(translate('Text File -> Tag', "Tag preview"))
        grid.addWidget(self.label, 0, 2)

        self.file = QTextEdit()
        grid.addWidget(self.file, 1, 0, 1, 2)

        self.tags = QTextEdit()
        grid.addWidget(self.tags, 1, 2, 1, 2)
        self.tags.setLineWrapMode(QTextEdit.LineWrapMode.NoWrap)

        hbox = QHBoxLayout()

        self.patterncombo = QComboBox()
        self.patterncombo.setEditable(True)
        self.patterncombo.setDuplicatesEnabled(False)

        okcancel = OKCancel()
        self.ok = okcancel.okButton
        self.cancel = okcancel.cancelButton

        self.openfile = QPushButton(
            translate('Text File -> Tag', "&Select File"))
        getclip = QPushButton(
            translate('Text File -> Tag', "&Paste Clipboard"))
        getclip.clicked.connect(self.openClipBoard)

        hbox.addWidget(self.openfile)
        hbox.addWidget(getclip)
        hbox.addWidget(self.patterncombo, 1)
        hbox.addLayout(okcancel)

        grid.addLayout(hbox, 3, 0, 1, 4)
        self.setLayout(grid)

        self.openfile.clicked.connect(self.openFile)
        self.cancel.clicked.connect(self.close)
        self.ok.clicked.connect(self.emitValues)

        if clipboard:
            self.openClipBoard()
            return

        self.lastDir = HOMEDIR

        if filename is not None:
            self.openFile(filename)

    def emitValues(self):
        """When I'm done, emit a signal with the updated tags."""
        self.close()
        self.Newtags.emit(self.dicttags,
                          str(self.patterncombo.currentText()))

    def fillTags(self, string=None):  # string is there purely for the SIGNAL
        """Fill the tag textbox."""

        def formattag(tags):
            if tags:
                return pprint_tag(tags, TAG_DISP, True)[:-2]
            else:
                return ""

        self.dicttags = []
        self.tags.clear()
        for z in self.lines.split("\n"):
            self.dicttags.append(findfunc.filenametotag(
                str(self.patterncombo.currentText()), z, False, False))
        if self.dicttags:
            self.tags.setHtml(
                "<br/>".join([formattag(z) for z in self.dicttags]))

    def openFile(self, filename=None, dirpath=None):
        """Open the file and fills the textboxes."""
        if not dirpath:
            dirpath = self.lastDir

        if not filename:
            selectedFile = QFileDialog.getOpenFileName(self,
                                                       'OpenFolder', dirpath)
            filename = selectedFile[0]

        if not filename:
            return True

        try:
            f = open(filename, 'r')
        except (IOError, OSError) as detail:
            errormsg = translate('Text File -> Tag',
                                 "The file <b>{}</b> couldn't be loaded.<br /> "
                                 "Do you want to choose another?"
                                 ).format(filename)

            ret = QMessageBox.question(self,
                                       translate('Text File -> Tag', "Error"),
                                       errormsg)

            if ret == QMessageBox.StandardButton.Yes:
                return self.openFile()
            else:
                return detail

        self.lines = f.readlines()
        self.file.setPlainText("".join(self.lines))
        self.setLines()
        self.fillTags()
        self.show()
        self.file.textChanged.connect(self.setLines)
        self.patterncombo.editTextChanged.connect(
            self.fillTags)
        self.lastDir = os.path.dirname(filename)

    def openClipBoard(self):
        text = str(QApplication.clipboard().text())
        self.lines = text.split('\n')
        self.file.setPlainText(text)
        self.setLines()
        self.fillTags()
        self.show()
        self.file.textChanged.connect(self.setLines)
        self.patterncombo.editTextChanged.connect(self.fillTags)

    def setLines(self):
        self.lines = str(self.file.document().toPlainText())
        self.fillTags()


class TextEdit(QPlainTextEdit):
    def focusInEvent(self, event):
        super(TextEdit, self).focusInEvent(event)
        self.selectAll()
        pos = len(self.toPlainText()) - 1
        if pos > 0:
            self.textCursor().setPosition(pos)
        else:
            self.textCursor().setPosition(0)

    def focusOutEvent(self, event):
        cursor = self.textCursor()
        cursor.clearSelection()
        self.setTextCursor(cursor)
        super(TextEdit, self).focusOutEvent(event)


class EditField(QDialog):
    """Dialog that allows editing of a field and it's values.

    When the user clicks ok, a 'donewithmyshit' (yeah yeah,
    I wrote this three years ago) signal
    is emitted containing three parameters.

    The first being the new field, the second that field's
    value and the third is the dictionary of the previous field
    in the form {field: value}.
    (Because the user might choose to edit a different tag,
    then the one that was chosen and you'd want to delete that one)"""
    donewithmyshit = pyqtSignal(str, str, object, name='donewithmyshit')

    def __init__(self, field=None, parent=None, field_list=None, edit=True):

        QDialog.__init__(self, parent)
        self.setWindowTitle(translate('Edit Field', 'Edit Field'))
        winsettings('edit_field', self)

        self.vbox = QVBoxLayout()

        label = QLabel(translate('Edit Field', "&Field"))
        self.tagcombo = QComboBox()
        self.tagcombo.setEditable(True)
        label.setBuddy(self.tagcombo)
        completer = self.tagcombo.completer()
        completer.setCaseSensitivity(Qt.CaseSensitivity.CaseSensitive)
        completer.setCompletionMode(QCompleter.CompletionMode.UnfilteredPopupCompletion)
        self.tagcombo.setCompleter(completer)
        self.tagcombo.addItems(field_list if field_list else gettaglist())

        # Get the previous field
        self.__oldField = field
        label1 = QLabel(translate('Edit Field', "&Value"))
        self.value = TextEdit()
        self.value.setTabChangesFocus(True)
        label1.setBuddy(self.value)

        okcancel = OKCancel()
        okcancel.okButton.setText(translate('Edit Field', 'A&dd'))

        if field is not None:
            x = self.tagcombo.findText(field[0])

            if x > -1:
                self.tagcombo.setCurrentIndex(x)
            else:
                self.tagcombo.setEditText(field[0])
            self.value.setPlainText(field[1])
            if edit:
                okcancel.okButton.setText(translate('Edit Field', 'E&dit'))

        list(map(self.vbox.addWidget, [label, self.tagcombo, label1, self.value]))

        self.vbox.addLayout(okcancel)
        self.setLayout(self.vbox)

        okcancel.ok.connect(self.ok)
        okcancel.cancel.connect(self.close)

        self.value.setFocus() if self.__oldField else self.tagcombo.setFocus()

    def ok(self):
        self.close()
        self.donewithmyshit.emit(
            str(self.tagcombo.currentText()),
            str(self.value.toPlainText()),
            self.__oldField)


class StatusWidgetItem(QTableWidgetItem):
    def __init__(self, text=None, status=None, colors=None, preview=False):
        QTableWidgetItem.__init__(self)
        self.preview = preview
        self.statusColors = colors

        if text:
            self.setText(text)

        if status and status in self.statusColors:
            self.setBackground(self.statusColors[status])

        self._status = status
        self.setFlags(Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled)
        self._original = (text, preview, status)
        self.linked = []

    @property
    def preview(self):
        return self.font().bold()

    @preview.setter
    def preview(self, value):
        font = self.font()
        font.setBold(value)
        self.setFont(font)

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        if status and status in self.statusColors:
            if self._status == ADD and status != REMOVE:
                return
            self.setBackground(self.statusColors[status])
            self._status = status
        else:
            self.setBackground(QTableWidgetItem().background())
            self._status = None

    def __lt__(self, item):
        if self.text().upper() < item.text().upper():
            return True
        return False

    def reset(self):
        self.setText(self._original[0])
        self.preview = self._original[1]
        status = self._original[2]
        if status == ADD:
            self.status = REMOVE
        else:
            self.status = status
        self.linked = []


class VerticalHeader(QHeaderView):
    def __init__(self, parent=None):
        QHeaderView.__init__(self, Qt.Orientation.Vertical, parent)
        self.setDefaultSectionSize(self.minimumSectionSize() + 4)
        self.setMinimumSectionSize(1)

        self.sectionResized.connect(
            self._resize)

    def _resize(self, row, oldsize, newsize):
        self.setDefaultSectionSize(newsize)


class StatusWidgetCombo(QComboBox):
    def __init__(self, items=None, status=None, colors=None, preview=False):
        QComboBox.__init__(self)
        self.preview = preview

        self.statusColors = colors
        self.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToContents)

        items = map(str, items)
        items = sorted(items, key=natural_sort_key)
        if len(items) > 1:
            items.append(r'\\'.join(items))
        self.addItem(KEEP)
        self.addItems(items)
        self.setCurrentIndex(0)

        if status and status in self.statusColors:
            self.setBackground(self.statusColors[status])
        else:
            self.setBackground(None)

        self._status = status
        self._original = (items, preview, status)
        self.linked = []

    @property
    def preview(self):
        return False

    @preview.setter
    def preview(self, _):
        return

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        if status and status in self.statusColors:
            if self._status == ADD and status != REMOVE:
                return
            self.setBackground(self.statusColors[status])
            self._status = status
        else:
            self.setBackground(None)
            self._status = None

    def setBackground(self, brush=None):
        if brush is None:
            color = QLineEdit().palette().color(QPalette.ColorRole.Base).name()
        else:
            color = brush.color().name()
        self.setStyleSheet("QComboBox { background-color: %s; }" % color)

    def background(self):
        brush = QBrush()
        brush.setColor(self.palette().color(QPalette.ColorRole.Window))
        return brush

    def reset(self):
        self.clear()
        self.addItems(self._original[0])
        self.setCurrentIndex(0)
        self.preview = self._original[1]
        status = self._original[2]
        if status == ADD:
            self.status = REMOVE
        else:
            self.status = status
        self.linked = []

    def setText(self, value):
        if self.currentIndex() == 0:
            self.insertItem(1, value)
            self.setCurrentIndex(1)
        else:
            self.setItemText(self.currentIndex(), value)


class ExTags(QDialog):
    """A dialog that shows you the tags in a file

    In addition, any attached cover art is shown."""
    rowChanged = pyqtSignal(object, name='rowChanged')
    extendedtags = pyqtSignal(dict, name='extendedtags')

    def __init__(self, parent=None, row=None, files=None, preview_mode=False,
                 artwork=True, status=None):

        if status is None:
            status = {'cover_pattern': 'folder'}

        self.status = status

        QDialog.__init__(self, parent)
        winsettings('extendedtags', self)
        self.get_fieldlist = []
        self.previewMode = preview_mode

        add = QColor.fromRgb(255, 255, 0)
        edit = QColor.fromRgb(0, 255, 0)
        remove = QColor.fromRgb(255, 0, 0)
        self._colors = {ADD: QBrush(add),
                        EDIT: QBrush(edit), REMOVE: QBrush(remove)}

        self.table = QTableWidget(0, 2, self)
        self.table.setVerticalHeader(VerticalHeader())
        self.table.verticalHeader().setVisible(False)
        self.table.setSortingEnabled(True)
        self.table.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.table.setHorizontalHeaderLabels([
            translate('Extended Tags', 'Field'),
            translate('Extended Tags', 'Value')])

        header = self.table.horizontalHeader()
        header.setVisible(True)
        header.setSortIndicatorShown(True)
        header.setStretchLastSection(True)
        header.setSortIndicator(0, Qt.SortOrder.AscendingOrder)

        self.piclabel = PicWidget(buttons=True)
        self.piclabel.imageChanged.connect(
            self._imageChanged)

        if not isinstance(self.piclabel.removepic, QAction):
            self.piclabel.removepic.clicked.connect(
                self.removePic)
        else:
            self.piclabel.removepic.triggered.connect(self.removePic)

        if row and row >= 0 and files:
            buttons = MoveButtons(files, row)
            buttons.indexChanged.connect(self._prevnext)
            buttons.setVisible(True)
        else:
            buttons = MoveButtons([], row)
            buttons.setVisible(False)

        self._files = files

        self.okcancel = OKCancel()
        self.okcancel.insertWidget(0, buttons)

        self._reset = QToolButton()
        self._reset.setToolTip(translate('Extended Tags',
                                         'Resets the selected fields to their original value.'))
        self._reset.setIcon(get_icon('edit-undo'))
        self._reset.clicked.connect(self.resetFields)

        self.listbuttons = ListButtons()
        self.listbuttons.layout().addWidget(self._reset)
        self.listbuttons.moveupButton.hide()
        self.listbuttons.movedownButton.hide()

        listframe = QFrame()
        listframe.setFrameStyle(QFrame.Shape.Box)
        hbox = QHBoxLayout()
        hbox.addWidget(self.table, 1)
        hbox.addLayout(self.listbuttons, 0)
        listframe.setLayout(hbox)

        layout = QVBoxLayout()
        if artwork:
            imageframe = QFrame()
            imageframe.setFrameStyle(QFrame.Shape.Box)
            vbox = QVBoxLayout()
            vbox.setContentsMargins(0, 0, 0, 0)
            vbox.addWidget(self.piclabel)
            vbox.addStretch()
            vbox.addStrut(0)
            imageframe.setLayout(vbox)

            hbox = QHBoxLayout()
            hbox.addWidget(listframe, 1)
            hbox.addSpacing(4)
            hbox.addWidget(imageframe)
            hbox.addStrut(1)
            layout.addLayout(hbox)
        else:
            layout.addWidget(listframe)

        layout.addLayout(self.okcancel)
        self.setLayout(layout)

        self.okcancel.cancel.connect(self.closeMe)
        self.table.itemDoubleClicked.connect(self.editField)
        self.table.itemSelectionChanged.connect(self._checkListBox)
        self.okcancel.ok.connect(self.okClicked)

        self.listbuttons.edit.connect(self.editField)
        self.listbuttons.addButton.clicked.connect(self.addField)
        self.listbuttons.removeButton.clicked.connect(self.removeField)
        self.listbuttons.duplicate.connect(self.duplicate)

        self.setMinimumSize(450, 350)

        self.canceled = False
        self.filechanged = False

        if row and row >= 0 and files:
            self._prevnext(row)
        else:
            self.loadFiles(files)

    def addField(self):
        win = EditField(parent=self, field_list=self.get_fieldlist)
        win.setModal(True)
        win.show()
        win.donewithmyshit.connect(self.editFieldBuddy)

    def _checkListBox(self):
        if self.table.rowCount() <= 0:
            self.table.setEnabled(False)
            self.listbuttons.editButton.setEnabled(False)
            self.listbuttons.removeButton.setEnabled(False)
            self.listbuttons.duplicateButton.setEnabled(False)
            self._reset.setEnabled(False)
        else:
            self.table.setEnabled(True)
            self._reset.setEnabled(True)
            if len(self.table.selectedIndexes()) / 2 > 1:
                self.listbuttons.editButton.setEnabled(False)
                self.listbuttons.duplicateButton.setEnabled(False)
            else:
                self.listbuttons.editButton.setEnabled(True)
                self.listbuttons.removeButton.setEnabled(True)
                self.listbuttons.duplicateButton.setEnabled(True)
        self.table.resizeColumnToContents(0)

    def closeEvent(self, event):
        self.piclabel.close()
        QDialog.closeEvent(self, event)

    def closeMe(self):
        self.canceled = True
        self.close()

    def _deletePressed(self, item):
        if self.table.deletePressed:
            self.table.deletePressed = False
            self.removeField()

    def duplicate(self):
        self.editField(True)

    def editField(self, duplicate=False):
        """Opens a dialog to edit the currently selected Field.

        If duplicate is True the Edit Field dialog will be populated
        with the currently selected field's values. The new field'll then
        be added to the field list."""
        row = self.table.currentRow()
        if row != -1:
            prevtag = self.get_field(row)
            if duplicate is True:
                win = EditField(prevtag, self, self.get_fieldlist, edit=False)
            else:
                win = EditField(prevtag, self, self.get_fieldlist)
            win.setModal(True)
            win.show()

            # Have to check for truth, because this method is
            # called by the doubleclicked signal.
            if duplicate is True:
                buddy = partial(self.editFieldBuddy, duplicate=True)
            else:
                buddy = self.editFieldBuddy
            win.donewithmyshit.connect(buddy)

    def editFieldBuddy(self, tag, value, prevtag=None, duplicate=False):
        rowcount = self.table.rowCount()
        if prevtag is not None:
            if duplicate:
                row = rowcount
                self._settag(rowcount, tag, value, ADD,
                             self.previewMode, True)
            else:
                if tag == prevtag[0]:
                    row = self.table.currentRow()
                    self._settag(row, tag, value, EDIT,
                                 self.previewMode, True)
                    if row + 1 < rowcount:
                        self.table.selectRow(row + 1)
                else:
                    cur_item = self.table.item(self.table.currentRow(), 0)
                    self.resetFields([cur_item])
                    self.table.setCurrentItem(cur_item,
                                              QItemSelectionModel.SelectionFlag.ClearAndSelect)
                    self.table.selectRow(self.table.row(cur_item))
                    self.removeField()
                    valitem = self._settag(rowcount, tag,
                                           value, ADD, self.previewMode, True)
                    cur_item.linked = [valitem]
        else:
            self._settag(rowcount, tag, value, ADD, self.previewMode, True)
        self._checkListBox()
        self.filechanged = True
        self.table.clearSelection()

    def get_field(self, row, status=None):
        getitem = self.table.item
        item = getitem(row, 0)
        tag = str(item.text())
        try:
            value = str(getitem(row, 1).text())
        except AttributeError:
            value = str(self.table.cellWidget(row, 1).currentText())
        if status:
            return (tag, value, item.status)
        else:
            return (tag, value)

    def _imageChanged(self):
        self.filechanged = True

    def loadSettings(self):
        cparser = PuddleConfig()
        self.get_fieldlist = gettaglist()
        get = lambda k, v: cparser.get('extendedtags', k, v, True)
        add = QColor.fromRgb(*get('add', [255, 255, 0]))
        edit = QColor.fromRgb(*get('edit', [0, 255, 0]))
        remove = QColor.fromRgb(*get('remove', [255, 0, 0]))

        self._colors = {ADD: QBrush(add),
                        EDIT: QBrush(edit), REMOVE: QBrush(remove)}

        item = self.table.item
        for row in range(self.table.rowCount()):
            field_item = self.get_item(row, 0)
            field_item.statusColors = self._colors
            field_item.status = field_item.status

            val_item = self.get_item(row, 1)
            val_item.statusColors = self._colors
            val_item.status = val_item.status

    def listtotag(self):
        get_field = self.get_field
        tags = {}
        lowered = {}
        listitems = [get_field(row, True) for row
                     in range(self.table.rowCount())]

        for field, val, status in listitems:
            if status != REMOVE:
                if val == KEEP:
                    continue
                l_field = field.lower()
                if l_field in lowered:
                    tags[lowered[l_field]].append(val)
                else:
                    lowered[l_field] = field
                    tags[field] = [z.strip() for z in val.split('\\') if z.strip()]
            else:
                if field.lower() not in lowered:
                    tags[field] = []
                    lowered[field.lower()] = field
        return tags

    def loadFiles(self, audios):
        if self.filechanged:
            self.save()
        self.filechanged = False
        self.table.clearContents()
        self.table.setRowCount(0)
        self.piclabel.lastfilename = audios[0].filepath
        self.piclabel.setEnabled(False)
        self.piclabel.setImages(None)

        if len(audios) == 1:
            audio = audios[0]
            self.setWindowTitle(audios[0].filepath)
            self._loadsingle(audio)
        else:
            self.setWindowTitle(
                translate('Extended Tags', 'Different files.'))

            from .tagmodel import status
            k = status['table'].model().taginfo[0]
            common, numvalues, imagetags = commontags(audios)
            images = common['__image']
            del (common['__image'])
            previews = set(audios[0].preview)
            italics = set(audios[0].equal_fields())
            self.piclabel.currentFile = audios[0]
            self.piclabel.filePattern = self.status['cover_pattern']

            for audio in audios[1:]:
                previews = previews.intersection(audio.preview)
                italics = italics.intersection(audio.equal_fields())

            row = 0

            for field, values in common.items():
                if field in italics:
                    preview = UNCHANGED
                # field in italics => field in previews.
                elif field in previews:
                    preview = BOLD
                else:
                    preview = UNCHANGED
                if numvalues[field] != len(audios):
                    self._settag(row, field, values, multi=True)
                    row += 1
                else:
                    if isinstance(values, str):
                        self._settag(row, field, values, None, preview)
                        row += 1
                    else:
                        for v in values:
                            self._settag(row, field, v, None, preview)
                            row += 1

            self.piclabel.setImageTags(imagetags)
            if images:
                self.piclabel.setEnabled(True)
                self.piclabel.setImages(images)
            else:
                self.piclabel.setImages(None)
                self.piclabel.setEnabled(True)
                if images == 0:
                    self.piclabel.context = 'Cover Varies'
                    self.piclabel.removepic.setEnabled(True)
        self._checkListBox()

    def _loadsingle(self, tags):
        items = []
        d = tags.usertags.copy()
        italics = tags.equal_fields()
        self.piclabel.currentFile = tags
        self.piclabel.filePattern = self.status['cover_pattern']

        for key, val in sorted(d.items()):
            if key in italics:
                preview = UNCHANGED
            elif key in tags.preview:
                preview = BOLD
            else:
                preview = UNCHANGED
            if isinstance(val, str):
                items.append([key, val, None, preview])
            else:
                [items.append([key, z, None, preview]) for z in val]

        [self._settag(i, *item) for i, item in enumerate(items)]

        self.piclabel.lastfilename = tags.filepath
        if not tags.library:
            self.piclabel.setImageTags(tags.IMAGETAGS)
            if tags.IMAGETAGS:
                if '__image' in tags.preview:
                    images = tags.preview['__image']
                else:
                    images = tags.images
                self.piclabel.setEnabled(True)
                if images:
                    self.piclabel.setImages(deepcopy(images))
                else:
                    self.piclabel.setImages(None)
        self._checkListBox()
        self.setWindowTitle(tags[PATH])

    def okClicked(self):
        self.save()
        self.close()

    def _prevnext(self, row):
        if self.filechanged:
            self.save()
        self.loadFiles([self._files[row]])
        self.rowChanged.emit(row)

    def get_item(self, row, column=None):
        if column is None:  # Assume QModelIndex passed
            column = row.column()
            row = row.row()
        item = self.table.item(row, column)
        if item is None:
            item = self.table.cellWidget(row, column)
        return item

    def removeField(self):
        tb = self.table
        tb.setSortingEnabled(False)
        to_remove = {}
        rows = []
        for index in self.table.selectedIndexes():
            row = index.row()
            item = self.get_item(index)
            if item.status == ADD:
                to_remove[row] = item
            rows.append(row)
            item.status = REMOVE
            item.status = REMOVE
        [tb.removeRow(tb.row(z)) for z in to_remove.values()]
        tb.setSortingEnabled(True)
        self.filechanged = True
        self._checkListBox()
        if rows:
            row = max(rows)
            self.table.clearSelection()
            if row + 1 < self.table.rowCount():
                self.table.selectRow(row + 1)

    def resetFields(self, items=None):
        box = self.table
        to_remove = {}  # Stores row: item values so that only one item
        # gets removed per row.

        max_row = -1
        for index in box.selectedIndexes():
            row = index.row()
            item = self.table.item(row, index.column())
            if item is None:
                item = self.table.cellWidget(row, index.column())
            for i in item.linked:
                try:
                    to_remove[box.row(i)] = i
                except RuntimeError:
                    pass
            item.reset()
            if row > max_row:
                max_row = row
            if item.status == REMOVE:
                to_remove[row] = item

        self.table.clearSelection()
        if max_row != -1 and max_row + 1 < self.table.rowCount():
            self.table.selectRow(max_row + 1)

        for item in to_remove.values():
            self.table.removeRow(self.table.row(item))
        self._checkListBox()

    def removePic(self):
        if self.piclabel.context == 'Cover Varies':
            self.piclabel.context = 'No Images'
            self.piclabel.removepic.setEnabled(False)
            if not isinstance(self.piclabel.removepic, QAction):
                self.piclabel.removepic.clicked.disconnect(
                    self.removePic)
            else:
                self.piclabel.removepic.triggered.disconnect(self.removePic)

            self.piclabel.setImages(None)

    def save(self):

        if not self.filechanged:
            table = self.table
            for row in range(table.rowCount()):
                combo = table.cellWidget(row, 1)
                if combo is not None and combo.currentIndex() != 0:
                    self.filechanged = True
                    break

        if not self.filechanged:
            return

        tags = self.listtotag()
        if self.piclabel.context != 'Cover Varies':
            if not self.piclabel.images:
                tags['__image'] = []
            else:
                tags["__image"] = self.piclabel.images
        newtags = [z for z in tags if z not in self.get_fieldlist]
        if newtags and newtags != ['__image']:
            settaglist(newtags + self.get_fieldlist)
        self.extendedtags.emit(tags)

    def _settag(self, row, field, value, status=None, preview=False,
                check=False, multi=False):

        tb = self.table
        tb.setSortingEnabled(False)
        if row >= tb.rowCount():
            tb.insertRow(row)
            field_item = StatusWidgetItem(field, status,
                                          self._colors, preview)
            tb.setItem(row, 0, field_item)
            if not multi and (len(value) == 1 or isinstance(value, str)):
                valitem = StatusWidgetItem(to_string(value), status, self._colors, preview)
                tb.setItem(row, 1, valitem)
            else:
                valitem = StatusWidgetCombo(value, status, self._colors, preview)
                tb.setCellWidget(row, 1, valitem)
        else:
            field_item = tb.item(row, 0)
            field_item.setText(field)
            field_item.status = status

            val_item = self.get_item(row, 1)
            val_item.setText(value)
            val_item.status = status

        if check:
            lowered_tag = field.lower()
            for row in range(tb.rowCount()):
                item = tb.item(row, 0)
                text = str(item.text())
                if text != field and text.lower() == lowered_tag:
                    item.setText(field)
                    if item.status not in [ADD, REMOVE]:
                        item.status = EDIT
                        try:
                            tb.item(row, 1).status = EDIT
                        except AttributeError:
                            tb.cellWidget(row, 1).status = EDIT

        tb.setSortingEnabled(True)
        return field_item


class ConfirmationErrorDialog(QDialog):
    def __init__(self, name, parent=None):
        super(ConfirmationErrorDialog, self).__init__(parent)
        self.name = name
        self.__label = QLabel()
        icon = QLabel()
        msgbox = QMessageBox()
        msgbox.setIcon(QMessageBox.Icon.Warning)
        icon.setPixmap(msgbox.iconPixmap())
        ok = QPushButton(translate("Defaults", "OK"))
        checkbox = QCheckBox(
            translate("Defaults", "Never show this message again."))

        ok.clicked.connect(partial(self.saveState, name))

        labelbox = QHBoxLayout()
        labelbox.addWidget(icon)
        labelbox.addWidget(self.__label)

        layout = QVBoxLayout()
        layout.addLayout(labelbox, 1)
        layout.addWidget(checkbox, 0, Qt.AlignmentFlag.AlignHCenter)
        layout.addWidget(ok, 0, Qt.AlignmentFlag.AlignHCenter)
        layout.addStretch()
        self.setLayout(layout)

    def saveState(self, name):
        settings = PuddleConfig()
        settings.set("OnceOnlyErrors", name, True)
        self.close()

    def showMessage(self, message):
        settings = PuddleConfig()
        should_show = settings.get("OnceOnlyErrors", self.name, True)
        self.__label.setText(message)
        if should_show:
            self.setModal(True)
            self.show()


if __name__ == "__main__":
    app = QApplication(sys.argv)
    dirpath = '/home/keith/Desktop/Daft.Punk-Random.Access.Memories-2013-FLAC.-NewsHost-1023'
    import glob

    tags = list(map(audioinfo.Tag, glob.glob(os.path.join(dirpath, "*.flac"))))
    for tag in tags:
        tag.preview = {}
        tag.equal_fields = lambda: []
        tag.library = None
    wid = ExTags(files=tags)
    wid.resize(200, 400)
    wid.show()
    pyqtRemoveInputHook()
    app.exec_()



from copy import deepcopy

from mutagen.mp4 import MP4, MP4Cover

from . import tag_versions, util
from .util import (usertags, isempty,
                   FILENAME, PATH,
                   getdeco, setdeco, FILETAGS, str_filesize, fn_hash, CaselessDict, keys_deco,
                   del_deco, cover_info, info_to_dict, parse_image, get_total, get_mime)

ATTRIBUTES = ('frequency', 'bitrate', 'length', 'accessed', 'size', 'created',
              'modified', 'bitspersample', 'channels')

# mp4 tags, like id3 can only have a fixed number of tags. The ones on the left
# with the corresponding tag as recognized by puddletag on the right...

TAGS = {
    '\xa9nam': 'title',
    '\xa9alb': 'album',
    '\xa9ART': 'artist',
    'aART': 'albumartist',
    '\xa9wrt': 'composer',
    '\xa9day': 'year',
    '\xa9cmt': 'comment',
    'desc': 'description',
    'purd': 'purchasedate',
    '\xa9grp': 'grouping',
    '\xa9gen': 'genre',
    '\xa9lyr': 'lyrics',
    'purl': 'podcasturl',
    'egid': 'podcastepisodeguid',
    'catg': 'podcastcategory',
    'keyw': 'podcastkeywords',
    '\xa9too': 'encodedby',
    'cprt': 'copyright',
    'soal': 'albumsortorder',
    'soaa': 'albumartistsortorder',
    'soar': 'artistsortorder',
    'sonm': 'titlesortorder',
    'soco': 'composersortorder',
    'sosn': 'showsortorder',
    'tvsh': 'showname',
    'cpil': 'partofcompilation',
    'pgap': 'partofgaplessalbum',
    'pcst': 'podcast',
    'tmpo': 'bpm',
    'akID': 'itunesaccounttype',
    'apID': 'itunesaccount',
    'atID': 'itunesartistid',
    'cmID': 'itunescomposerid',
    'cnID': 'itunescatalogid',
    'geID': 'itunesgenreid',
    'plID': 'itunesalbumid',
    'prID': 'itunesproductid',
    'sfID': 'itunesstorecountry',
    'rtng': 'itunesadvisory',
    'stik': 'itunesmediatype',
    'xid ': 'itunesxid',
    '\xa9pub': 'itunespublisher',
}

REVTAGS = dict([reversed(z) for z in TAGS.items()])

# Because these values are storted in different ways, text for album, tuple
# for tracks, I created to a bunch of functions, to handle the reading
# and writing of these.

# Functions are in get and set pairs.
# get functions take a value mutagen expects it and returns it in puddletag's format.
# set functions do the opposite.

encode = lambda x: [z.encode('utf8') for z in x]


def getbool(value):
    if value:
        return ['Yes']
    else:
        return ['No']


def setbool(value):
    if value == 'No':
        return False
    elif value:
        return True
    else:
        return False


def settext(text):
    if isinstance(text, str):
        return [str(text)]
    elif isinstance(text, str):
        return [text]
    else:
        return [str(z) for z in text]


def gettext(value):
    return value


def settuple(value):
    temp = []
    for tup in value:
        if isinstance(tup, str):
            values = [z.strip() for z in tup.split('/')]
            try:
                temp.append(tuple([int(z) for z in values][:2]))
            except (TypeError, IndexError):
                continue
        else:
            temp.append(tup)
    return temp


def gettuple(value):
    return [str(track) + '/' + str(total) for track, total in value]


def getint(value):
    return [str(z) for z in value]


def setint(value):
    if isinstance(value, (int, str)):
        return [int(value)]
    temp = []
    for z in value:
        try:
            temp.append(int(z))
        except ValueError:
            continue
    return temp


FUNCS = {
    'title': (gettext, settext),
    'album': (gettext, settext),
    'artist': (gettext, settext),
    'albumartist': (gettext, settext),
    'composer': (gettext, settext),
    'year': (gettext, settext),
    'comment': (gettext, settext),
    'description': (gettext, settext),
    'purchasedate': (gettext, settext),
    'grouping': (gettext, settext),
    'genre': (gettext, settext),
    'lyrics': (gettext, settext),
    'podcasturl': (gettext, settext),
    'podcastepisodeguid': (gettext, settext),
    'podcastcategory': (gettext, settext),
    'podcastkeywords': (gettext, settext),
    'encodedby': (gettext, settext),
    'copyright': (gettext, settext),
    'albumsortorder': (gettext, settext),
    'albumartistsortorder': (gettext, settext),
    'artistsortorder': (gettext, settext),
    'titlesortorder': (gettext, settext),
    'composersortorder': (gettext, settext),
    'showsortorder': (gettext, settext),
    'showname': (gettext, settext),
    'partofcompilation': (getbool, setbool),
    'partofgaplessalbum': (getbool, setbool),
    'podcast': (getbool, setbool),
    'track': (getint, setint),
    'disc': (getint, setint),
    'totaltracks': (getint, setint),
    'totaldiscs': (getint, setint),
    'bpm': (getint, setint),
    'itunesaccounttype': (getint, setint),
    'itunesaccount': (gettext, settext),
    'itunesartistid': (getint, setint),
    'itunescomposerid': (gettext, settext),
    'itunescatalogid': (getint, setint),
    'itunesgenreid': (getint, setint),
    'itunesalbumid': (getint, setint),
    'itunesproductid': (gettext, settext),
    'itunesstorecountry': (getint, setint),
    'itunesadvisory': (getint, setint),
    'itunesmediatype': (getint, setint),
    'itunesxid': (gettext, settext),
    'itunespublisher': (gettext, settext),
}


def bin_to_pic(cover):
    try:
        format = cover.imageformat
    except AttributeError:
        format = cover.format
    if format == MP4Cover.FORMAT_PNG:
        return {'data': cover, 'mime': 'image/png'}
    else:
        return {'data': cover, 'mime': 'image/jpeg'}


def pic_to_bin(image):
    data = image[util.DATA]
    mime = get_mime(data)
    if mime == 'image/png':
        format = MP4Cover.FORMAT_PNG
    elif mime == 'image/jpeg':
        format = MP4Cover.FORMAT_JPEG
    else:
        return
    return MP4Cover(data, format)


class Tag(util.MockTag):
    """Class for AAC tags."""

    mapping = {}
    revmapping = {}
    IMAGETAGS = (util.MIMETYPE, util.DATA)

    def __init__(self, filename=None):
        self.__images = []
        self.__errors = set()
        self.__tags = CaselessDict()

        util.MockTag.__init__(self, filename)

    @property
    def filepath(self):
        return util.MockTag.get_filepath(self)

    @filepath.setter
    def filepath(self, val):
        self.__tags.update(util.MockTag.set_filepath(self, val))

    def __contains__(self, key):
        if key == '__image':
            return bool(self.images)

        elif key == '__total':
            try:
                return bool(get_total(self))
            except (KeyError, ValueError):
                return False

        if self.revmapping:
            key = self.revmapping.get(key, key)
        return key in self.__tags

    def __deepcopy__(self, memo=None):
        cls = Tag()
        cls.mapping = self.mapping
        cls.revmapping = self.revmapping
        cls.set_fundamentals(deepcopy(self.__tags), deepcopy(self.images),
                             self.mut_obj, deepcopy(self.__freeform), deepcopy(self.__errors))
        cls.filepath = self.filepath
        return cls

    @del_deco
    def __delitem__(self, key):
        if key == '__image':
            self.images = []
        elif key.startswith('__'):
            return
        else:
            del (self.__tags[key])

    @getdeco
    def __getitem__(self, key):
        if key.startswith('__'):
            if key == '__image':
                return self.images
            elif key == '__total':
                return FUNCS['totaltracks'][0](self.__tags['totaltracks'])
            else:
                return self.__tags[key]

        try:
            return FUNCS[key][0](self.__tags[key])
        except KeyError:
            return gettext(self.__tags[key])

    @setdeco
    def __setitem__(self, key, value):
        if isinstance(key, int):
            self.__tags[key] = value
            return
        elif key.startswith('__'):
            if key == '__image':
                self.images = value
            if key in FILETAGS:
                setattr(self, fn_hash[key], value)
            elif key == '__total':
                self.__tags['totaltracks'] = FUNCS['totaltracks'][1](value)
            return
        elif isempty(value):
            if key in self:
                del (self[key])
            return
        else:
            try:
                new_val = FUNCS[key][1](value)
                if value:
                    self.__tags[key] = new_val
            except KeyError:
                # User defined tags.
                self.__freeform[key] = '----:com.apple.iTunes:%s' % key
                self.__tags[key] = settext(value)
            except ValueError:
                pass

    def delete(self):
        self.mut_obj.delete()
        for key in self.usertags:
            del (self.__tags[self.revmapping.get(key, key)])
        self.images = []

    @property
    def images(self):
        return self.__images

    @images.setter
    def images(self, images):
        if images:
            self.__images = [parse_image(i, self.IMAGETAGS) for i in images]
        else:
            self.__images = []
        cover_info(images, self.__tags)

    @property
    def info(self):
        info = self.mut_obj.info
        fileinfo = [('Path', self[PATH]),
                    ('Size', str_filesize(int(self.size))),
                    ('Filename', self[FILENAME]),
                    ('Modified', self.modified)]

        mp4info = [('Bitrate', self.bitrate),
                   ('Frequency', self.frequency),
                   ('Channels', str(info.channels)),
                   ('Length', self.length),
                   ('Bits per sample', str(info.bits_per_sample))]

        return [('File', fileinfo), ('MP4 Info', mp4info)]

    def link(self, filename):
        """Links the audio, filename
        returns self if successful, None otherwise."""

        tags, audio = self.load(filename, MP4)
        self.images = []

        if audio is None:
            return

        revmap, mapping = self.revmapping, self.mapping
        self.revmapping, self.mapping = {}, {}

        self.__freeform = {}  # Keys are tags as required by mutagen i.e. The '----'
        # frames. Values are the tag as represented by puddletag.

        if audio.tags:  # Not empty
            keys = list(audio.keys())
            try:
                self.images = list(map(bin_to_pic, audio['covr']))
                keys.remove('covr')
            except KeyError:
                self.images = []

            convert = lambda k, v: FUNCS[k][1](v)

            # I want 'trkn', to split into track and totaltracks, like Mp3tag.
            if 'trkn' in keys:
                tags['track'] = convert('track',
                                        [z[0] for z in audio['trkn']])
                tags['totaltracks'] = convert('totaltracks',
                                              [z[1] for z in audio['trkn']])
                keys.remove('trkn')

            # Same as above
            if 'disk' in keys:
                tags['disc'] = convert('disc', [z[0] for z in audio['disk']])
                tags['totaldiscs'] = convert('totaldiscs',
                                             [z[1] for z in audio['disk']])
                keys.remove('disk')

            for key in keys:
                if key in TAGS:
                    tags[TAGS[key]] = convert(TAGS[key], audio[key])
                else:
                    field = key[key.find(':', key.find(':') + 1) + 1:]

                    self.__freeform[field] = key
                    try:
                        field_value = []
                        for v in audio[key]:
                            if isinstance(v, bytes):
                                field_value.append(str(v, 'utf8'))
                            elif isinstance(v, str):
                                field_value.append(v)
                            else:
                                field_value.append(str(v))
                        tags[field] = field_value
                    except UnicodeDecodeError:
                        self.__errors.add(field)

        for k, v in list(tags.items()):
            if not v:
                del (tags[k])

        self.__tags.update(info_to_dict(audio.info))
        self.__tags.update(tags)
        self.revmapping, self.mapping = revmap, mapping
        self.__tags['__tag_read'] = 'MP4'
        self.filetype = 'MP4'
        self.__tags['__filetype'] = self.filetype
        self.set_attrs(ATTRIBUTES, self.__tags)
        self.update_tag_list()
        self.mut_obj = audio

    @keys_deco
    def keys(self):
        return list(self.__tags.keys())

    def save(self):
        if self.mut_obj.tags is None:
            self.mut_obj.add_tags()
        if self.filepath != self.mut_obj.filename:
            self.mut_obj.filename = self.filepath
        audio = self.mut_obj

        newtag = {}
        tuples = (('track', ['trkn', 'totaltracks']),
                  ('disc', ['disk', 'totaldiscs']))
        tags = self.__tags
        for tag, values in tuples:
            if tag in tags:
                denom = tags[tag]
                if values[1] in tags:
                    total = tags[values[1]]
                    newtag[values[0]] = [(int(t), int(total)) for t, total in
                                         zip(denom, total)]
                else:
                    newtag[values[0]] = [(int(z), 0) for z in denom]
            elif values[1] in tags:
                total = tags[values[1]]
                newtag[values[0]] = [(0, int(z)) for z in total]

        tags = usertags(self.__tags)
        tags = [(z, tags[z]) for z in tags
                if z not in ['track', 'totaltracks', 'disc', 'totaldiscs']]

        for tag, value in tags:
            try:
                newtag[REVTAGS[tag]] = value
            except KeyError:
                newtag[self.__freeform[tag]] = encode(self.__tags[tag])

        if self.images:
            newtag['covr'] = [_f for _f in map(pic_to_bin, self.images) if _f]

        toremove = [z for z in audio.keys() if
                    z not in newtag and z not in self.__errors]
        for key in toremove:
            del (audio[key])
        audio.update(newtag)
        audio.save()

    def set_fundamentals(self, tags, images, mut_obj, freeform=None, errors=None):
        self.__freeform = {} if freeform is None else freeform
        self.__errors = {} if errors is None else errors
        self.__tags = tags
        self.mut_obj = mut_obj
        self.images = images

    def update_tag_list(self):
        l = tag_versions.tags_in_file(self.filepath)
        if l:
            self.__tags['__tag'] = 'MP4, ' + ', '.join(l)
        else:
            self.__tags['__tag'] = 'MP4'


filetype = (MP4, Tag, 'MP4', ['m4a', 'mp4', 'm4v'])



import logging
import os
import sys
from os.path import exists
from importlib import import_module

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication, QDialog, QHBoxLayout, QLabel, QListWidget, QListWidgetItem, QVBoxLayout
from configobj import ConfigObj

from .constants import (FUNCTIONS, TAGSOURCE, SAVEDIR, PLUGINDIR,
                        DIALOGS, MUSICLIBS, MODULES, FUNCTIONS_NO_PREVIEW)
from .puddleobjects import PuddleConfig, winsettings
from .translations import translate

NAME = 'name'
AUTHOR = 'author'
DESC = 'description'
PT_VERSION = 'puddletag_version'
VERSION = 'version'
INFO_SECTION = 'info'
MODULE_NAME = 'module'

PLUGIN_DIRS = [PLUGINDIR,
               os.path.join(os.path.dirname(__file__), 'plugins')]

PROPERTIES = [NAME, AUTHOR, DESC, PT_VERSION, VERSION]


def get_plugins(plugindir):
    if not os.path.exists(plugindir):
        os.makedirs(plugindir)
    infos = []
    for module in os.listdir(plugindir):
        info_path = os.path.join(plugindir, module, 'info')
        if not exists(info_path):
            continue
        config_data = dict(ConfigObj(info_path))
        values = [config_data.get(INFO_SECTION, {}).get(prop, '') for prop in PROPERTIES]

        if len([_f for _f in values if _f]) < len(PROPERTIES):
            continue
        d = dict(list(zip(PROPERTIES, values)))
        d[MODULE_NAME] = module
        infos.append(d)
    return infos


def load_plugins(plugins=None, parent=None):
    [sys.path.insert(0, d) for d in PLUGIN_DIRS]
    cparser = PuddleConfig()
    to_load = cparser.get('plugins', 'to_load', [])
    functions = {}
    tagsources = []
    dialogs = []
    musiclibs = []
    modules = []
    functions_no_preview = []

    join = os.path.join
    if plugins is None:
        plugins = []
        [plugins.extend(get_plugins(d)) for d in PLUGIN_DIRS]

    plugins.sort(key=lambda d: d.get(NAME, ''))

    plugin_actions = {
        'functions': functions.update,
        'functions_no_preview':  functions_no_preview.extend,
        'tagsources':  tagsources.extend,
        'dialogs':  dialogs.extend,
        'musiclibs':  musiclibs.extend
        }

    for plugin in plugins:
        if plugin[MODULE_NAME] in to_load:
            try:
                module = import_module('puddlestuff.plugins.' + plugin[MODULE_NAME])
                for attribute, action in plugin_actions.items():
                    if hasattr(module, attribute):
                        action(getattr(module, attribute))
                modules.append(module)
            except Exception as e:
                logging.exception(f"Failed to load plugin {plugin[NAME]}; error={e}")                

    for d in PLUGIN_DIRS:
        del (sys.path[0])

    return {FUNCTIONS: functions, TAGSOURCE: tagsources, DIALOGS: dialogs,
            MUSICLIBS: musiclibs, MODULES: modules, FUNCTIONS_NO_PREVIEW: functions_no_preview}


class InfoWidget(QLabel):
    def __init__(self, info=None, parent=None):
        super(InfoWidget, self).__init__(parent)
        self.setAlignment(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignTop)
        self.setWordWrap(True)
        if info:
            self.changeInfo(info)

    def changeInfo(self, info):
        labels = [translate('Plugin Settings', 'Name'),
                  translate('Plugin Settings', 'Author'),
                  translate('Plugin Settings', 'Description'),
                  translate('Plugin Settings', 'Version')]
        properties = [NAME, AUTHOR, DESC, VERSION]

        text = '<br />'.join(['<b>%s:</b> %s' % (disp, info[prop]) for
                              disp, prop in zip(labels, properties)])
        self.setText(text)


class PluginConfig(QDialog):
    def __init__(self, parent=None):
        super(PluginConfig, self).__init__(parent)
        winsettings('pluginconfig', self)
        self._listbox = QListWidget()
        info_display = InfoWidget()

        hbox = QHBoxLayout()
        hbox.addWidget(self._listbox, 0)
        hbox.addWidget(info_display, 1)

        vbox = QVBoxLayout()
        vbox.addLayout(hbox)
        vbox.addWidget(
            QLabel(translate("Plugin Settings",
                             '<b>Loading/unloading plugins requires a restart.</b>')))
        self.setLayout(vbox)

        plugins = []
        [plugins.extend(get_plugins(d)) for d in PLUGIN_DIRS]

        cparser = PuddleConfig()
        to_load = cparser.get('plugins', 'to_load', [])
        plugins.sort(key=lambda d: d.get(NAME, ''))
        for plugin in plugins:
            item = QListWidgetItem()
            item.setText(plugin[NAME])
            if plugin[MODULE_NAME] in to_load:
                item.setCheckState(Qt.CheckState.Checked)
            else:
                item.setCheckState(Qt.CheckState.Unchecked)
            item.plugin = plugin
            self._listbox.addItem(item)

        self._listbox.currentItemChanged.connect(
            lambda item, previous: info_display.changeInfo(item.plugin))

    def get_to_load(self):
        to_load = []
        for row in range(self._listbox.count()):
            item = self._listbox.item(row)
            if item.checkState() == Qt.CheckState.Checked:
                to_load.append(item.plugin[MODULE_NAME])
        return to_load

    def applySettings(self, control=None):
        to_load = self.get_to_load()
        cparser = PuddleConfig()
        cparser.set('plugins', 'to_load', to_load)


if __name__ == '__main__':
    app = QApplication([])
    win = PluginConfig()
    win.show()
    app.exec_()



# -*- coding: utf-8 -*-
import os

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QHBoxLayout, QListWidgetItem, QWidget

from . import audioinfo
from .constants import CONFIGDIR
from .puddleobjects import (ListButtons, ListBox)


def load_genres(filepath=None):
    if not filepath:
        filepath = os.path.join(CONFIGDIR, 'genres')
    try:
        return [x.strip() for x in open(filepath, 'r').readlines() if x.strip()]
    except (IOError, OSError):
        return audioinfo.GENRES[::]


def save_genres(genres, filepath=None):
    if not filepath:
        filepath = os.path.join(CONFIGDIR, 'genres')
    f = open(filepath, 'w')
    f.write('\n'.join(genres))
    f.close()


class Genres(QWidget):
    def __init__(self, parent=None, status=None):
        QWidget.__init__(self, parent)
        if status is None:
            self._status = {}
            genres = load_genres()
        else:
            self._status = status
            genres = status['genres']

        self.listbox = ListBox()
        self._itemflags = Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEditable | Qt.ItemFlag.ItemIsEnabled
        [self.listbox.addItem(self._createItem(z)) for z in genres]

        buttons = ListButtons()
        self.listbox.connectToListButtons(buttons)
        self.listbox.setAutoScroll(False)

        buttons.add.connect(self.add)
        buttons.edit.connect(self.edit)

        hbox = QHBoxLayout()
        hbox.addWidget(self.listbox, 1)
        hbox.addLayout(buttons, 0)
        self.setLayout(hbox)

    def _createItem(self, text):
        item = QListWidgetItem(text)
        item.setFlags(self._itemflags)
        return item

    def add(self):
        self.listbox.setAutoScroll(True)
        item = self._createItem('')
        self.listbox.addItem(item)
        self.listbox.clearSelection()
        self.listbox.setCurrentItem(item)
        self.listbox.editItem(item)
        self.listbox.setAutoScroll(False)

    def edit(self):
        self.listbox.setAutoScroll(True)
        self.listbox.editItem(self.listbox.currentItem())
        self.listbox.setAutoScroll(False)

    def applySettings(self, control=None):
        item = self.listbox.item
        genres = [str(item(row).text()) for row in
                  range(self.listbox.count())]
        self._status['genres'] = genres



# -*- coding: utf-8 -*-
import os
from functools import partial

from PyQt5.QtCore import QFileSystemWatcher, QItemSelectionModel, pyqtSignal
from PyQt5.QtWidgets import QAction, QApplication, QDialog, QHBoxLayout, QLabel, QLineEdit, QListWidgetItem, QPushButton, QVBoxLayout

from . import puddleobjects as puddleobjects
from . import puddletag
from .constants import ACTIONDIR
from .findfunc import load_macro_info as load_action
from .puddleobjects import (create_buddy, ListBox,
                            ListButtons, load_actions, OKCancel, PuddleConfig)
from .translations import translate

FILENAME = os.path.join(ACTIONDIR, 'action_shortcuts')

NAME = 'name'
FILENAMES = 'filenames'
SHORTCUT_SECTION = 'Shortcut'


def create_action_shortcut(name, filenames, scut_key=None, method=None, parent=None, add=False):
    if not method:
        from .mainwin.funcs import applyaction
        method = applyaction

    if not parent:
        parent = puddletag.status['mainwin']

    shortcut = Shortcut(name, filenames, method, parent, scut_key)

    if add:
        from .puddletag import add_shortcuts
        add_shortcuts('&Actions', [shortcut], save=bool(scut_key))
    return shortcut


def create_action_shortcuts(method, parent=None):
    actions, shortcuts = load_settings()
    menu_shortcuts = []
    for name, filenames in shortcuts:
        menu_shortcuts.append(Shortcut(name, filenames, method, parent))
    return menu_shortcuts


def get_shortcuts(default=None):
    from .puddletag import status
    if status['actions']:
        ret = [_f for _f in (str(z.shortcut().toString()) for z in status['actions']) if _f]
    else:
        ret = []
    if default:
        return set(ret + default)
    else:
        return set(ret)


def load_settings(filename=None, actions=None):
    if filename is None:
        filename = FILENAME

    if not os.path.exists(os.path.dirname(filename)):
        os.makedirs(os.path.dirname(filename))

    cparser = PuddleConfig(filename)

    actions = load_actions() if actions is None else actions

    shortcuts = []
    for section in sorted(cparser.sections()):
        if section.startswith(SHORTCUT_SECTION):
            name = cparser.get(section, NAME, 'Default')
            filenames = cparser.get(section, FILENAMES, [])
            shortcuts.append([name, filenames])
    return actions, shortcuts


def save_shortcut(name, filenames):
    cparser = PuddleConfig(FILENAME)
    section = SHORTCUT_SECTION + str(len(cparser.sections()))
    cparser.set(section, NAME, name)
    cparser.set(section, FILENAMES, filenames)


class Shortcut(QAction):
    def __init__(self, name, filenames, method, parent, shortcut=''):
        super(Shortcut, self).__init__(name, parent)

        self.enabled = 'filesselected'
        self.control = None
        self.command = None
        self.togglecheck = None
        self._method = method

        if shortcut:
            self.setShortcut(shortcut)
        self.filenames = filenames
        self.funcs = self.get_funcs()
        self.triggered.connect(self.runAction)

        self._watcher = QFileSystemWatcher(list(filter(os.path.exists, filenames)),
                                           self)
        self._watcher.fileChanged.connect(self._checkFile)

    def _checkFile(self, filename):
        # There's some fucked up behaviour going on with QFileSystemWatcher.
        # Without the code here, this method will be called 404 times
        # whenever the file changes. I spent the last hour trying
        # to figure it out and I'm giving up and brute forcing it.

        # Disconnecting and reconnecting doesn't work either.
        # Nor does using blockSignals

        self._watcher.fileChanged.disconnect(self._checkFile)
        filename = filename
        if not os.path.exists(filename):
            self.filenames.remove(filename)
        self.funcs = self.get_funcs()
        self._watcher = QFileSystemWatcher(self.filenames, self)
        self._watcher.fileChanged.connect(self._checkFile)

    def get_funcs(self, filenames=None):
        if filenames is None:
            filenames = self.filenames
        funcs = []
        for f in filenames:
            if os.path.exists(f):
                funcs.extend(load_action(f)[0])
        return funcs

    def runAction(self):
        return self._method(funcs=self.funcs)


class Editor(QDialog):
    actionChanged = pyqtSignal(str, list, str, name='actionChanged')

    def __init__(self, title='Add Action', shortcut='', actions=None, names=None, shortcuts=None, parent=None):
        super(Editor, self).__init__(parent)
        self.setWindowTitle(title)

        self._items = {}

        self._name = QLineEdit('Name')

        if shortcut and shortcut in shortcuts:
            shortcuts.remove(shortcut)

        self._shortcut = puddleobjects.ShortcutEditor(shortcuts)
        self._shortcut.setText(shortcut)
        clear = QPushButton(translate('Shortcuts', '&Clear'))
        clear.clicked.connect(self._shortcut.clear)

        if names is None:
            names = []
        self._names = names

        self._actionList = ListBox()
        self._actionList.itemDoubleClicked.connect(self._addAction)
        self._newActionList = ListBox()
        listbuttons = ListButtons()
        listbuttons.duplicateButton.hide()
        listbuttons.insertStretch(0)
        listbuttons.add.connect(self._addAction)

        self._newActionList.connectToListButtons(listbuttons)

        okcancel = OKCancel()
        okcancel.ok.connect(self.okClicked)
        okcancel.cancel.connect(self.close)
        self._ok = okcancel.okButton
        self._name.textChanged.connect(self.enableOk)
        scut_status = QLabel('')
        self._shortcut.validityChanged.connect(
            lambda v: scut_status.setText('') if v or (not self._shortcut.text()) else
            scut_status.setText(translate('Shortcuts', "Invalid shortcut sequence.")))
        okcancel.insertWidget(0, scut_status)

        hbox = QHBoxLayout()
        hbox.addLayout(
            create_buddy('Actions', self._actionList, QVBoxLayout()), 1)
        hbox.addLayout(listbuttons, 0)
        hbox.addLayout(create_buddy('Actions to run for shortcut',
                                    self._newActionList, QVBoxLayout()), 1)

        layout = QVBoxLayout()
        layout.addLayout(create_buddy('Shortcut &Name: ', self._name))
        scut_layout = create_buddy('&Keyboard Shortcut: ', self._shortcut)
        scut_layout.addWidget(clear)
        layout.addLayout(scut_layout)
        layout.addLayout(hbox)
        layout.addLayout(okcancel)
        self.setLayout(layout)

        if actions:
            self.setActions(actions)

    def _addAction(self, item=None):
        if item is None:
            for item in self._actionList.selectedItems():
                self._addAction(item)
            return
        new_item = QListWidgetItem(item)
        new_item._action = item._action
        self._newActionList.addItem(new_item)
        self._newActionList.setCurrentItem(new_item,
                                           QItemSelectionModel.SelectionFlag.ClearAndSelect)

    def enableOk(self, text):
        if not text or text in self._names:
            self._ok.setEnabled(False)
        else:
            self._ok.setEnabled(True)

    def okClicked(self):
        alist = self._newActionList
        items = list(map(alist.item, range(alist.count())))
        actions = [item._action[1] for item in items]
        self.actionChanged.emit(str(self._name.text()), actions,
                                str(self._shortcut.text()) if self._shortcut.valid else '')
        self.close()

    def setActions(self, actions):
        self._actionList.clear()
        self._actions = []
        for funcs, name, filename in actions:
            item = QListWidgetItem(name)
            item.setToolTip('\n'.join([func.description() for func in funcs]))
            item._action = [name, filename]
            self._actionList.addItem(item)

    def setName(self, name):
        self._name.setText(name)

    def setAttrs(self, name, actions, filenames, shortcut=''):
        names = dict([(z[2], z[1]) for z in actions])
        self.setActions(actions)
        self.setName(name)
        self._newActionList.clear()
        self.setShortcut(shortcut)
        if filenames:
            for filename in filenames:
                item = QListWidgetItem(names.get(filename, translate('Shortcuts', '(Deleted)')))
                item._action = [names.get(filename, ''), filename]
                self._newActionList.addItem(item)

    def setShortcut(self, text):
        self._shortcut.setText(text)


class ShortcutEditor(QDialog):
    def __init__(self, load=False, parent=None, buttons=False):
        super(ShortcutEditor, self).__init__(parent)
        self._names = []
        self._hotkeys = []

        self._listbox = ListBox()

        listbuttons = ListButtons()
        listbuttons.insertStretch(0)
        listbuttons.add.connect(self._addShortcut)
        listbuttons.edit.connect(self._editShortcut)
        listbuttons.duplicate.connect(self._duplicate)
        self._listbox.itemDoubleClicked.connect(self._editShortcut)

        self._listbox.connectToListButtons(listbuttons)

        hbox = QHBoxLayout()
        hbox.addLayout(create_buddy('Shortcuts', self._listbox, QVBoxLayout()))
        hbox.addLayout(listbuttons)

        okcancel = OKCancel()

        vbox = QVBoxLayout()
        vbox.addLayout(hbox)
        if buttons:
            vbox.addLayout(okcancel)
        self.setLayout(vbox)

        okcancel.ok.connect(self.okClicked)
        okcancel.cancel.connect(self.close)

        if load:
            self.loadSettings()

    def _addShortcut(self):
        shortcuts = get_shortcuts().difference(self._hotkeys).union(
            i.shortcut for i in self._listbox.items() if i.shortcut)

        win = Editor('Add Shortcut', '', self._actions, self.names(), shortcuts, self)
        win.setModal(True)
        win.actionChanged.connect(self.addShortcut)
        win.show()

    def addShortcut(self, name, filenames, shortcut='', select=True):
        item = QListWidgetItem(name)
        item.actionName = name
        item.filenames = filenames[::]
        item.shortcut = shortcut
        self._listbox.addItem(item)
        if select:
            self._listbox.setCurrentItem(item, QItemSelectionModel.SelectionFlag.ClearAndSelect)

    def applySettings(self, control=None):
        from .puddletag import remove_shortcuts, add_shortcuts
        remove_shortcuts('&Actions', self._names)

        f = open(FILENAME, 'w')
        f.close()

        cparser = PuddleConfig(FILENAME)
        for i, item in enumerate(self._listbox.items()):
            section = SHORTCUT_SECTION + str(i)
            cparser.set(section, NAME, item.actionName)
            cparser.set(section, FILENAMES, item.filenames)

        from .mainwin.funcs import applyaction

        shortcuts = create_action_shortcuts(applyaction, control)
        for item, shortcut in zip(list(self._listbox.items()), shortcuts):
            if item.shortcut:
                shortcut.setShortcut(item.shortcut)
        add_shortcuts('&Actions', shortcuts, save=True)

    def okClicked(self):
        self.applySettings()
        self.close()

    def _duplicate(self):
        try:
            item = self._listbox.selectedItems()[0]
        except IndexError:
            return
        shortcuts = get_shortcuts().difference(self._hotkeys).union(
            i.shortcut for i in self._listbox.items() if i.shortcut)
        win = Editor('Duplicate Shortcut', '', self._actions, self.names(), shortcuts, self)
        win.setAttrs(item.actionName, self._actions, item.filenames, '')
        win.setModal(True)
        win.actionChanged.connect(self.addShortcut)
        win.show()

    def _editShortcut(self):
        try:
            item = self._listbox.selectedItems()[0]
        except IndexError:
            return
        shortcuts = get_shortcuts().difference(self._hotkeys).union(
            i.shortcut for i in self._listbox.items() if i.shortcut)

        names = self.names()
        names.remove(item.actionName)
        win = Editor('Edit Shortcut', item.shortcut, self._actions, names, shortcuts, self)
        win.setAttrs(item.actionName, self._actions, item.filenames, item.shortcut)
        win.setModal(True)
        win.actionChanged.connect(
            partial(self.editShortcut, item))
        win.show()

    def editShortcut(self, item, name, filenames, shortcut):
        item.actionName = name
        item.filenames = filenames[::]
        item.shortcut = shortcut
        item.setText(name)

    def loadSettings(self, filename=None, actions=None):
        self._names = []
        self._hotkeys = []

        if filename is None:
            filename = os.path.join(ACTIONDIR, 'action_shortcuts')

        self._listbox.clear()
        cparser = PuddleConfig(filename)

        if actions is None:
            self._actions = load_actions()
        else:
            self._actions = actions

        from .puddletag import status
        if status['actions']:
            shortcuts = dict((str(a.text()), str(a.shortcut().toString()))
                             for a in status['actions'])
        else:
            shortcuts = {}

        for section in sorted(cparser.sections()):
            if section.startswith('Shortcut'):
                name = cparser.get(section, NAME, 'Default')
                self._names.append(name)
                filenames = cparser.get(section, FILENAMES, [])
                shortcut = shortcuts.get(name, '')
                self.addShortcut(name, filenames, shortcut, select=False)
                self._hotkeys.append(shortcut)

    def names(self):
        return [item.actionName for item in self._listbox.items()]


if __name__ == '__main__':
    app = QApplication([])
    actions = load_actions()
    win = ShortcutEditor(buttons=True)
    win.loadSettings()
    win.show()
    app.exec_()



import logging
import os
import re
import socket
import urllib.error
import urllib.parse
import urllib.request
from html.parser import HTMLParser
from importlib import import_module
from os.path import join, exists

from PyQt5.QtCore import QObject, pyqtSignal

from .. import version_string
from ..constants import CONFIGDIR
from ..findfunc import tagtofilename
from ..puddleobjects import PuddleConfig
from ..util import translate


class FoundEncoding(Exception):
    pass


class WebServiceError(EnvironmentError):
    pass


class RetrievalError(WebServiceError):

    def __init__(self, msg, code=0):
        WebServiceError.__init__(self, msg)
        self.code = code


class SubmissionError(WebServiceError):

    def __init__(self, msg, code=0):
        WebServiceError.__init__(self, msg)
        self.code = code


class _SignalObject(QObject):
    statusChanged = pyqtSignal(str, name='statusChanged')
    logappend = pyqtSignal(str, name='logappend')


cparser = PuddleConfig()

COVERDIR = cparser.get('tagsources', 'coverdir', join(CONFIGDIR, 'covers'))
COVER_PATTERN = '%artist% - %album%'
SAVECOVERS = False

mapping = {}
status_obj = _SignalObject()
useragent = "puddletag/" + version_string


def get_encoding(page, decode=False, default=None):
    encoding = None
    match = re.search(r'<\?xml(.+)\?>', page)
    if match:
        enc = re.search(r'''encoding(?:\s*)=(?:\s*)["'](.+?)['"]''',
                        match.group(), re.I)
        if enc:
            encoding = enc.groups()[0]

    if not encoding and default:
        encoding = default

    return encoding


def find_id(tracks, field):
    if not field:
        return
    for track in tracks:
        if field in track:
            value = track[field]
            if isinstance(value, str):
                return value
            else:
                return value[0]


# From http://stackoverflow.com/questions/4389572 by bobince
def iri_to_uri(iri):
    parts = urllib.parse.urlparse(iri)
    return urllib.parse.urlunparse(
        part.encode('idna') if i == 1
        else part.encode('utf-8')
        for i, part in enumerate(parts)
    ).decode()


def parse_searchstring(text):
    try:
        text = [z.split(';') for z in text.split('|') if z]
        return [(z.strip(), v.strip()) for z, v in text]
    except ValueError:
        raise RetrievalError(
            translate('Tag Sources',
                      '<b>Error parsing artist/album combinations.</b>'))


def retrieve_cover(url):
    write_log(translate('Tag Sources', "Retrieving cover: {}").format(url))
    cover = urlopen(url)
    return {'__image': [{'data': cover}]}


def save_cover(info, data, filetype):
    filename = tagtofilename(COVER_PATTERN, info, True, filetype)
    save_file(filename, data)


def save_file(filename, data):
    path = join(filename, COVERDIR)
    if exists(path):
        save_file('%s0' % filename)
        return
    f = open(filename, 'wb')
    f.write(data)
    f.close()


def set_coverdir(dirpath):
    global COVERDIR
    COVERDIR = dirpath


def set_savecovers(value):
    global SAVECOVERS
    SAVECOVERS = value


def set_mapping(m):
    global mapping

    mapping.clear()
    mapping.update(m)


def set_status(msg):
    status_obj.statusChanged.emit(msg)


def get_useragent():
    if useragent:
        return useragent
    else:
        return 'puddetag/' + version_string


def set_useragent(agent):
    global useragent
    useragent = agent


def to_file(data, name):
    if os.path.exists(name):
        return to_file(data, name + '_')

    f = open(name, 'w')
    f.write(data)
    f.close()


def url_encode_non_ascii(b):
    return re.sub('[\x80-\xFF]', lambda c: '%%%02x' % ord(c.group(0)), b)


def urlopen(url, mask=True, code=False):
    try:
        request = urllib.request.Request(url)
        if useragent:
            request.add_header('User-Agent', useragent)
        page = urllib.request.build_opener().open(request)
        if page.code == 403:
            raise RetrievalError(
                translate("Tag Sources", 'HTTPError 403: Forbidden'))
        elif page.code == 404:
            raise RetrievalError(
                translate("Tag Sources", "Page doesn't exist"))
        if code:
            return page.read(), page.code
        else:
            return page.read()
    except urllib.error.URLError as e:
        try:
            msg = '%s (%s)' % (e.reason.strerror, e.reason.errno)
        except AttributeError:
            msg = str(e)

        try:
            raise RetrievalError(msg, e.code)
        except AttributeError:
            msg = e.args[1] if len(e.args) > 1 else e.args[0]
            raise RetrievalError(
                translate('Defaults', "Connection Error: {}").format(msg))
    except socket.error as e:
        msg = '%s (%s)' % (e.strerror, e.code)
        raise RetrievalError(msg)
    except EnvironmentError as e:
        msg = '%s (%s)' % (e.strerror, e.code)
        raise RetrievalError(msg)


def write_log(text):
    status_obj.logappend.emit(text)


class MetaProcessor(HTMLParser):

    def reset(self):
        self.pieces = []
        self.encoding = None
        HTMLParser.reset(self)

    def start_meta(self, text):
        text = [tuple([x.lower() for x in z]) for z in text]
        if text[0] == ('http-equiv', 'content-type'):
            d = dict(text)
            if 'charset' in d:
                encoding = d['charset']
                error = FoundEncoding()
                error.encoding = encoding
                raise error
            if text[1][0] == 'content':
                encoding = re.search('charset.*=(.+)', text[1][1]).group(1)
                error = FoundEncoding()
                error.encoding = encoding
                raise error


tagsources = []
for source in ('acoust_id', 'amazon', 'amg', 'discogs', 'freedb',
               'musicbrainz'):
    try:
        tagsources.append(getattr(
            import_module('puddlestuff.tagsources.' + source),
            'info'))
    except ImportError as ie:
        logging.debug(f"error loading source '{source}: error={ie}")



import os
import sip
import sys

from PyQt5.QtCore import QEvent, QThread, Qt, pyqtRemoveInputHook, pyqtSignal
from PyQt5.QtGui import QBrush
from PyQt5.QtWidgets import QApplication, QComboBox, QCompleter, QDialog, QGroupBox, QHBoxLayout, \
    QLabel, QLineEdit, QTableWidget, QTableWidgetItem, QVBoxLayout, QWidget

from ..audioinfo import INFOTAGS
from ..puddleobjects import ListButtons, PuddleConfig

pyqtRemoveInputHook()
from ..constants import LEFTDOCK, SELECTIONCHANGED, BLANK, KEEP, SEPARATOR
from functools import partial
from ..puddlesettings import SettingsError
from ..translations import translate
from ..constants import CONFIGDIR


def loadsettings(filepath=None):
    settings = PuddleConfig()
    if filepath:
        settings.filename = filepath
    else:
        settings.filename = os.path.join(CONFIGDIR, 'tagpanel')
    numrows = settings.get('panel', 'numrows', -1, True)
    if numrows > -1:
        sections = settings.sections()
        d = {}
        for row in range(numrows):
            section = str(row)
            tags = settings.get(section, 'tags', [''])
            titles = settings.get(section, 'titles', [''])
            d[row] = list(zip(titles, tags))
    else:
        titles = ['&Artist', '&Title', 'Al&bum', 'T&rack', '&Year', "&Genre", '&Comment']
        tags = ['artist', 'title', 'album', 'track', 'year', 'genre', 'comment']
        newtags = list(zip(titles, tags))
        d = {0: [newtags[0]], 1: [newtags[1]], 2: [newtags[2]],
             3: [newtags[3], newtags[4], newtags[5]],
             4: [newtags[6]]}
    return d


def savesettings(d, filepath=None):
    settings = PuddleConfig()
    if filepath:
        settings.filename = filepath
    else:
        settings.filename = os.path.join(settings.savedir, 'tagpanel')
    settings.set('panel', 'numrows', str(len(d)))
    for row, rowtags in d.items():
        settings.set(str(row), 'tags', [z[1] for z in rowtags])
        settings.set(str(row), 'titles', [z[0] for z in rowtags])


class Combo(QComboBox):

    def __init__(self, *args, **kwargs):
        super(Combo, self).__init__(*args, **kwargs)
        self._edited = False

        def k(text): self._edited = True

        self.editTextChanged.connect(k)

    def setEditText(self, text):
        self._edited = False
        super(Combo, self).setEditText(text)

    def focusOutEvent(self, event):
        if not self._edited:
            return super(Combo, self).focusOutEvent(event)
        curtext = self.currentText()
        index = self.findText(curtext,
                              Qt.MatchFlag.MatchExactly | Qt.MatchFlag.MatchFixedString | Qt.MatchFlag.MatchCaseSensitive)
        if index == -1:
            index = self.findText(curtext, Qt.MatchFlag.MatchExactly | Qt.MatchFlag.MatchFixedString)
        if index > 1:
            if curtext == self.itemText(index):
                self.removeItem(index)
            self.insertItem(2, curtext)
        self._edited = False
        return super(Combo, self).focusOutEvent(event)


class Thread(QThread):
    update = pyqtSignal(object, name='update')

    def __init__(self, values, parent=None):
        self._values = values
        QThread.__init__(self, parent)

    def run(self):
        self.update.emit(self._values)


class FrameCombo(QGroupBox):
    """A group box with combos that allow to edit
    tags individually if so inclined.

    tags should be a list with the tags
    that each combo box should hold specified
    in the form [(Display Value, internal tag)]
    .e.g [("Artist","artist"), ("Track Number", "track")]

    Individual comboboxes can be accessed by using FrameCombo.combos
    which is a dictionary key = tag, value = respective combobox.
    """

    onetomany = pyqtSignal(dict, name='onetomany')
    onetomanypreview = pyqtSignal(dict, name='onetomanypreview')
    manypreview = pyqtSignal(list, name='manypreview')

    def __init__(self, tags=None, parent=None, status=None):
        self.settingsdialog = SettingsWin
        QGroupBox.__init__(self, parent)
        self.emits = ['onetomany', 'onetomanypreview', 'manypreview']
        self.receives = [(SELECTIONCHANGED, self.fillCombos),
                         ('previewModeChanged', lambda v: self._enablePreview()
                         if v else self._disablePreview()), ]
        self.combos = {}
        self.labels = {}
        self._hboxes = []
        self._status = status
        self._originalValues = {}
        self.__indexFuncs = []

    def _disablePreview(self):
        self.__disconnectIndexChanged
        self.__indexFuncs = []
        for field, combo in self.combos.items():
            edit = QLineEdit()
            combo.setLineEdit(edit)
            completer = combo.completer()
            completer.setCaseSensitivity(Qt.CaseSensitivity.CaseSensitive)
            completer.setCompletionMode(QCompleter.CompletionMode.UnfilteredPopupCompletion)

    def disableCombos(self):
        for z in self.combos:
            self.combos[z].clear()
            self.combos[z].setEnabled(False)

    def _enablePreview(self):
        self.__indexFuncs = []
        for field, combo in self.combos.items():
            func = partial(self._emitChange, field)
            edit = QLineEdit()
            combo.setLineEdit(edit)
            completer = combo.completer()
            completer.setCaseSensitivity(Qt.CaseSensitivity.CaseSensitive)
            completer.setCompletionMode(QCompleter.CompletionMode.UnfilteredPopupCompletion)
            edit.textEdited.connect(func)
            combo.currentIndexChanged.connect(func)
            self.__indexFuncs.append((combo, func))
            edit.editingFinished.connect(self.save)

    def __disconnectIndexChanged(self):
        for combo, func in self.__indexFuncs:
            combo.disconnect(combo, func)

    def _emitChange(self, field, text):
        if type(text) is int:
            text = self.combos[field].itemText(text)
        else:
            text = str(text)

        if text == BLANK:
            text = ''
        elif text == KEEP:
            if self._audios:
                self.manypreview.emit([{field: a.get(field, '')} for a in self._audios])
        else:
            if field in INFOTAGS:
                value = text
            else:
                value = text.split(SEPARATOR)
        self.onetomanypreview.emit({field: text})

    def fillCombos(self, *args):
        audios = self._status['selectedfiles']
        self._audios = [z.usertags for z in audios]
        combos = self.combos

        if not audios:
            for combo in combos.values():
                combo.clear()
                combo.setEnabled(False)
            return

        [combo.blockSignals(True) for combo in combos.values()]
        self.initCombos(True)
        tags = dict((tag, set()) for tag in combos)
        for audio in audios:
            for field in tags:
                if field in audio:
                    value = audio[field]
                    if isinstance(value, str):
                        tags[field].add(value)
                    else:
                        tags[field].add(SEPARATOR.join(value))
                else:
                    tags[field].add('')

        for field, values in tags.items():
            combo = combos[field]
            combo.addItems(sorted(values))
            if len(values) == 1:
                combo.setCurrentIndex(2)
            else:
                combo.setCurrentIndex(0)

        if 'genre' in tags and len(tags['genre']) == 1:
            combo = combos['genre']
            values = sorted(tags['genre'])
            index = combo.findText(values[0])
            if index > -1:
                combo.setCurrentIndex(index)
            else:
                combo.setEditText(values[0])
        elif 'genre' in tags:
            combos['genre'].setCurrentIndex(0)

        self._originalValues = dict([(field, str(combo.currentText()))
                                     for field, combo in self.combos.items()])
        self._originalValues['__image'] = self._status['images']
        [combo.blockSignals(False) for combo in combos.values()]

    def save(self):
        """Writes the tags of the selected files to the values in self.combogroup.combos."""
        combos = self.combos
        tags = {}

        images = self._status['images']
        if images is not None:
            tags['__image'] = images
        originals = {}
        for field, combo in combos.items():
            curtext = str(combo.currentText())
            if self._originalValues[field] == curtext:
                continue
            originals[field] = curtext
            if curtext == BLANK:
                tags[field] = []
            elif curtext == KEEP:
                continue
            else:
                if field in INFOTAGS:
                    tags[field] = curtext
                else:
                    tags[field] = curtext.split(SEPARATOR)

        if '__image' in tags:
            if self._originalValues['__image'] == tags['__image']:
                del (tags['__image'])
            else:
                originals['__image'] = tags['__image']

        if not tags:
            return

        self._originalValues.update(originals)
        self.onetomany.emit(tags)
        if 'genre' in combos:
            combo = combos['genre']

            genres = self._status['genres']
            new_genres = [_f for _f in str(combo.currentText()).split(SEPARATOR) if _f]

            [genres.append(genre) for genre in new_genres
             if genre not in genres]

    def setCombos(self, rowtags):
        """Creates a vertical column of comboboxes.
        tags are tags is usual in the (tag, backgroundvalue) case[should be enumerable].
        rows are a dictionary with each key being a list of the indexes of tags that should
        be one one row.

        E.g Say you wanted to have the artist, album, title, and comments on
        seperate rows. But then you want year, genre and track on one row. With that
        row being before the comments row. You'd do something like...

        >>>tags = [('Artist', 'artist'), ('Title', 'title'), ('Album', 'album'),
        ...        ('Track', 'track'), ("Comments",'comment'), ('Genre', 'genre'), ('Year', 'date')]
        >>>rows = {0:[0], 1:[1], 2:[2], 3[3,4,6],4:[5]
        >>>f = FrameCombo()
        >>>f.setCombo(tags, rows)"""
        # pdb.set_trace()

        if self.combos:
            vbox = self.layout()
            for box, control in self._hboxes:
                box.removeWidget(control)
                sip.delete(control)
                if not box.count():
                    vbox.removeItem(box)
                    sip.delete(box)
        else:
            vbox = QVBoxLayout()
            vbox.setContentsMargins(0, 0, 0, 0)
        self.combos = {}
        self.labels = {}
        self._hboxes = []

        j = 0
        for row, tags in sorted(rowtags.items()):
            labelbox = QHBoxLayout()
            labelbox.setContentsMargins(6, 1, 1, 1)
            widgetbox = QHBoxLayout()
            widgetbox.setContentsMargins(0, 0, 0, 0)
            for tag in tags:
                tagval = tag[1]
                self.labels[tagval] = QLabel(tag[0])
                self.combos[tagval] = QComboBox(self)
                self.combos[tagval].setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
                self.combos[tagval].setEditable(True)
                self.combos[tagval].completer().setCompletionMode(
                    QCompleter.CompletionMode.UnfilteredPopupCompletion)
                self.combos[tagval].completer().setCaseSensitivity(
                    Qt.CaseSensitivity.CaseSensitive)
                self.labels[tagval].setBuddy(self.combos[tagval])
                labelbox.addWidget(self.labels[tagval])
                widgetbox.addWidget(self.combos[tagval])
                self._hboxes.append((labelbox, self.labels[tagval]))
                self._hboxes.append((widgetbox, self.combos[tagval]))
            vbox.addLayout(labelbox)
            vbox.addLayout(widgetbox)

        self.setLayout(vbox)
        QApplication.processEvents()
        self.setMaximumHeight(self.sizeHint().height())

    def eventFilter(self, obj, event):
        if isinstance(obj, QComboBox) and event.type() == QEvent.Type.FocusOut:
            return False
        return QGroupBox.eventFilter(self, obj, event)

    def initCombos(self, enable=False):
        """Clears the comboboxes and adds two items, <keep> and <blank>.
        If <keep> is selected and the tags in the combos are saved,
        then they remain unchanged. If <blank> is selected, then the tag is removed"""
        for combo in self.combos.values():
            combo.clear()
            combo.setEnabled(enable)
            combo.addItems([KEEP, BLANK])

        if 'genre' in self.combos:
            pass
            self.combos['genre'].addItems(self._status['genres'])

    def reloadCombos(self, tags):
        self.setCombos(tags)

    def loadSettings(self):
        self.setCombos(loadsettings())


class PuddleTable(QTableWidget):
    def __init__(self, columns=None, defaultvals=None, parent=None):
        QTableWidget.__init__(self, parent)
        self._default = defaultvals
        self.verticalHeader().hide()
        if columns:
            self.setColumnCount(len(columns))
            self.setHorizontalHeaderLabels(columns)
            self.horizontalHeader().setStretchLastSection(True)

    def add(self, texts=None):
        row = self.rowCount()
        item = self.horizontalHeaderItem
        if self._default and (not texts):
            texts = self._default
        elif not texts:
            texts = [item(z).text() for z in range(self.columnCount())]
        self.insertRow(row)
        for column, v in enumerate(texts):
            self.setItem(row, column, QTableWidgetItem(v))

    def remove(self):
        rows = self.selectedRows()
        if not rows:
            return
        for i in range(len(rows)):
            self.removeRow(rows[i])
            rows = [z - 1 for z in rows]
        self.itemSelectionChanged.emit()

    def moveUp(self):
        row = self.currentRow()
        if row <= 0:
            return
        previtems = self.texts(row - 1)
        newitems = self.texts(row)
        self.setRowTexts(row, previtems)
        self.setRowTexts(row - 1, newitems)
        curitem = self.item(row - 1, self.currentItem().column())
        self.setCurrentItem(curitem)

    def moveDown(self):
        row = self.currentRow()
        if row >= self.rowCount() - 1:
            return
        previtems = self.texts(row + 1)
        newitems = self.texts(row)
        self.setRowTexts(row, previtems)
        self.setRowTexts(row + 1, newitems)
        curitem = self.item(row + 1, self.currentItem().column())
        self.setCurrentItem(curitem)

    @property
    def rows(self):
        return range(self.rowCount())

    def setRowTexts(self, row, texts):
        item = self.item
        for column, text in zip(list(range(self.columnCount())), texts):
            item(row, column).setText(text)

    def texts(self, row):
        item = self.item
        return [str(item(row, z).text()) for z in range(self.columnCount())]

    def items(self, row):
        item = self.item
        return [item(row, z) for z in range(self.columnCount())]

    def text(self, row, column):
        return str(self.item(row, column).text())

    def selectedRows(self):
        return sorted(set(i.row() for i in self.selectedIndexes()))


TABLEWIDGETBG = QTableWidgetItem().background()
RED = QBrush(Qt.GlobalColor.red)

TITLE = translate("Defaults", 'Title')
FIELD = translate("Defaults", 'Field')
ROW = translate("Tag Panel Settings", 'Row')


class SettingsWin(QWidget):

    def __init__(self, parent=None, status=None):
        QDialog.__init__(self, parent)
        self.title = translate('Settings', 'Tag Panel')
        self._table = PuddleTable([TITLE, FIELD, ROW],
                                  [TITLE, FIELD, '0'], self)
        self._buttons = ListButtons()
        self._buttons.connectToWidget(self._table, add=self.add,
                                      edit=self.edit, moveup=self._table.moveUp,
                                      movedown=self._table.moveDown, duplicate=self.duplicate)

        hbox = QHBoxLayout()
        hbox.addWidget(self._table, 1)
        hbox.addLayout(self._buttons, 0)
        self.setLayout(hbox)

        self._table.cellChanged.connect(self._checkItem)
        self._table.itemSelectionChanged.connect(self._enableButtons)
        self.fill()

    def add(self, texts=None):
        table = self._table
        if not texts:
            text = table.text
            rows = []
            for row in table.rows:
                try:
                    rows.append(int(text(row, 2)))
                except (TypeError, ValueError):
                    pass
            row = str(max(rows) + 1) if rows else '1'
            table.add([TITLE, FIELD.lower(), row])
        else:
            table.add(texts)
        item = table.item(table.rowCount() - 1, 0)
        table.setCurrentItem(item)
        table.editItem(item)

    def fill(self):
        d = loadsettings()
        self._old = d
        for row in d:
            for z in d[row]:
                self._table.add(z + (str(row),))

    def applySettings(self, control=None):
        texts = self._table.texts
        d = {}
        for row in self._table.rows:
            l = texts(row)
            try:
                l[2] = int(l[2])
            except ValueError:
                raise SettingsError(translate('Tag Panel Settings',
                                              'All row numbers must be integers.'))
            try:
                d[l[2]].append(l[:-1])
            except KeyError:
                d[l[2]] = [l[:-1]]
        d = dict([(i, d[v]) for i, v in enumerate(sorted(d))])  # consecutive rows
        if self._old == d:
            return
        savesettings(d)
        control.setCombos(d)

    def edit(self):
        self._table.editItem(self._table.currentItem())

    def _checkItem(self, row, column):
        table = self._table
        if column == 2:
            try:
                int(table.text(row, column))
                table.item(row, column).setBackground(TABLEWIDGETBG)
            except ValueError:
                i = table.item(row, column)
                i.setBackground(RED)

    def _enableButtons(self):
        table = self._table
        if table.rowCount() <= 0:
            self._buttons.editButton.setEnabled(False)
            self._buttons.duplicateButton.setEnabled(False)
            self._buttons.removeButton.setEnabled(False)
        elif self._table.selectedRows():
            self._buttons.editButton.setEnabled(True)
            self._buttons.duplicateButton.setEnabled(True)
            self._buttons.removeButton.setEnabled(True)

    def duplicate(self):
        table = self._table
        row = table.currentRow()
        if row < 0: return
        self.add(table.texts(row))


control = ('Tag Panel', FrameCombo, LEFTDOCK, True)

if __name__ == '__main__':
    app = QApplication(sys.argv)
    win = SettingsWin()
    win.show()
    app.exec_()
